#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Jul  2 09:55:18 2024

@author: thahn
"""

# =============================================================================
# This file contains the functions used to calculate statistics and data related to ARM products
# =============================================================================

import numpy as np
import xarray as xr
from tqdm import tqdm
from vincenty import vincenty

from .calculate_convective_properties import (
    calculate_interp_sonde_convective_properties,
)


# Calculate nearest item in list to given pivot
def find_nearest(array: np.ndarray, pivot) -> int:
    array = np.asarray(array)
    idx = (np.abs(array - pivot)).argmin()
    return idx


def extract_arm_product(
    analysis_object: dict,
    path_to_files: str,
    variable_names: list[str] | str,
    **args: dict,
) -> xr.Dataset:
    """


    Parameters
    ----------
    analysis_object : dict
        A CoCoMET-US standard analysis object containing at least US_tracks.
    path_to_files : str
        A glob-like path to the ARM product output.
    variable_name : list[str] | str
        Case sensitive name of variable you want to extract from the ARM data

    Returns
    -------
    output_data : xarray.core.dataset.Dataset
        An xarray Dataset with the following: frame, tracking_time, arm_time, time_delta, closest_feature_id (km), variable_names list

    """

    # Ensure input is a list
    if not isinstance(variable_names, list):
        variable_names = [variable_names]

    # Open vap product
    vap = xr.open_mfdataset(
        path_to_files, coords="all", concat_dim="time", combine="nested"
    )

    vap_info = {
        "frame": [],
        "tracking_time": [],
        "arm_time": [],
        "time_delta": [],
        "closest_feature_id": [],
        "closest_cell_id": [],
        "distance_to_closest_feature": [],
    }

    for var in variable_names:
        vap_info[var] = []

    frame_groups = analysis_object["US_tracks"].groupby("frame")

    # If 1D input data, don't use heights
    if len(vap.dims) == 1:
        # Loop over frames
        for ii, frame in tqdm(
            enumerate(frame_groups),
            desc="=====Extracting ARM Data=====",
            total=frame_groups.ngroups,
        ):
            # Get VAP at current time step
            time_idx = find_nearest(vap.time.values, frame[1].time.values[0])
            time_delta = vap.time.values[time_idx] - frame[1].time.values[0]

            # Get the position of the device
            lat_pos = vap.lat.values[time_idx]
            lon_pos = vap.lon.values[time_idx]

            feature_id = []
            cell_distance = []

            # Loop over features
            for feature in frame[1].groupby("feature_id"):
                # Get distance from each feature to the device
                dis_to_vdis = vincenty(
                    (lat_pos, lon_pos),
                    (feature[1].latitude.values, feature[1].longitude.values),
                )
                cell_distance.append(dis_to_vdis)
                feature_id.append(feature[0])

            closest_feature_id = feature_id[
                np.where(cell_distance == np.nanmin(cell_distance))[0][0]
            ]
            closest_feature = frame[1].query("feature_id==@closest_feature_id")

            vap_info["frame"].append(frame[0])
            vap_info["tracking_time"].append(frame[1].time.values[0])
            vap_info["arm_time"].append(vap.time.values[time_idx])
            vap_info["time_delta"].append(time_delta)
            vap_info["closest_feature_id"].append(closest_feature.feature_id.values[0])
            vap_info["closest_cell_id"].append(closest_feature.cell_id.values[0])
            vap_info["distance_to_closest_feature"].append(np.nanmin(cell_distance))

            for var in variable_names:
                vap_info[var].append(vap[var].values[time_idx])

        data_vars = {
            "time_delta": ("time", vap_info["time_delta"]),
            "closest_feature_id": ("time", vap_info["closest_feature_id"]),
            "closest_cell_id": ("time", vap_info["closest_cell_id"]),
            "distance_to_closest_feature": (
                "time",
                vap_info["distance_to_closest_feature"],
            ),
        }

        for var in variable_names:
            data_vars[var] = ("time", vap_info[var])

        # Create output Dataset
        output_data = xr.Dataset(
            coords={
                "frame": ("time", vap_info["frame"]),
                "tracking_time": ("time", vap_info["tracking_time"]),
                "arm_time": ("time", vap_info["arm_time"]),
            },
            data_vars=data_vars,
            attrs={
                "description": f"ARM Data {vap.input_datastreams}, Variables: {variable_names}"
            },
        )

        return output_data

    # If 2D data, add heights
    # Loop over frames
    for ii, frame in tqdm(
        enumerate(frame_groups),
        desc="=====Extracting ARM Data=====",
        total=frame_groups.ngroups,
    ):
        # Get VAP at current time step
        time_idx = find_nearest(vap.time.values, frame[1].time.values[0])
        time_delta = vap.time.values[time_idx] - frame[1].time.values[0]

        # Get the position of the device
        lat_pos = vap.lat.values[time_idx]
        lon_pos = vap.lon.values[time_idx]

        feature_id = []
        cell_distance = []

        # Loop over features
        for feature in frame[1].groupby("feature_id"):
            # Get distance from each feature to the device
            dis_to_vdis = vincenty(
                (lat_pos, lon_pos),
                (feature[1].latitude.values, feature[1].longitude.values),
            )
            cell_distance.append(dis_to_vdis)
            feature_id.append(feature[0])

        closest_feature_id = feature_id[
            np.where(cell_distance == np.nanmin(cell_distance))[0][0]
        ]
        closest_feature = frame[1].query("feature_id==@closest_feature_id")

        vap_info["frame"].append(frame[0])
        vap_info["tracking_time"].append(frame[1].time.values[0])
        vap_info["arm_time"].append(vap.time.values[time_idx])
        vap_info["time_delta"].append(time_delta)
        vap_info["closest_feature_id"].append(closest_feature.feature_id.values[0])
        vap_info["closest_cell_id"].append(closest_feature.cell_id.values[0])
        vap_info["distance_to_closest_feature"].append(np.nanmin(cell_distance))

        for var in variable_names:
            vap_info[var].append(vap[var].values[time_idx, :])

    data_vars = {
        "time_delta": ("time", vap_info["time_delta"]),
        "closest_feature_id": ("time", vap_info["closest_feature_id"]),
        "closest_cell_id": ("time", vap_info["closest_cell_id"]),
        "distance_to_closest_feature": (
            "time",
            vap_info["distance_to_closest_feature"],
        ),
    }

    for var in variable_names:
        data_vars[var] = (["time", "height"], vap_info[var])

    # Create output Dataset
    output_data = xr.Dataset(
        coords={
            "frame": ("time", vap_info["frame"]),
            "tracking_time": ("time", vap_info["tracking_time"]),
            "arm_time": ("time", vap_info["arm_time"]),
            "height": ((vap.height.values) * 1000 - vap.alt.values.min()),
        },
        data_vars=data_vars,
        attrs={
            "description": f"ARM Data {vap.input_datastreams}, Variables: {variable_names}"
        },
    )

    return output_data


def calculate_convective_indices(
    analysis_object: dict, path_to_files: str, **args: dict
) -> tuple[xr.Dataset, xr.Dataset]:
    """


    Parameters
    ----------
    analysis_object : dict
        A CoCoMET-US standard analysis object containing at least US_tracks.
    path_to_files : str
        A glob-like path to the INTERPSONDE ARM product output.
    **args : dict
        Parameters to pass to the calculations of convective initation properties..

    Returns
    -------
    sonde_output_indices_data : xarray.core.dataset.Dataset
        An xarray Dataset with the following: sonde_time, ...

    """

    # Open video disdrometer product
    sonde = xr.open_mfdataset(
        path_to_files, coords="all", concat_dim="time", combine="nested"
    )

    convective_init_info = {
        "sonde_time": [],
        "cell_id": [],
        "cape": [],
        "ncape": [],
        "cin": [],
        "lnb": [],
        "lfc": [],
        "lcl": [],
        "wind_shear": [],
        "low_rh": [],
        "mid_rh": [],
        "richardson": [],
        "elr_0_3": [],
    }

    init_groups = (
        analysis_object["US_tracks"].query("lifetime_percent==0").groupby("cell_id")
    )

    # Loop over cell initations
    for ii, cell in tqdm(
        enumerate(init_groups),
        desc="=====Calculating INTERPSONDE Initation Properties=====",
        total=init_groups.ngroups,
    ):
        # Get sonde data at time step before initiation if possible
        time_idx = find_nearest(sonde.time.values, cell[1].time.values[0])
        time_idx_init = time_idx - 1 if time_idx > 0 else 0

        # Calculate CAPE, CIN, etc.
        properties = calculate_interp_sonde_convective_properties(
            sonde.isel(time=time_idx_init), **args
        )

        convective_init_info["sonde_time"].append(sonde.time.values[time_idx_init])
        convective_init_info["cell_id"].append(cell[0])
        convective_init_info["cape"].append(properties["CAPE"])
        convective_init_info["ncape"].append(properties["NCAPE"])
        convective_init_info["cin"].append(properties["CIN"])
        convective_init_info["lnb"].append(properties["LNB"])
        convective_init_info["lfc"].append(properties["LFC"])
        convective_init_info["lcl"].append(properties["LCL"])
        convective_init_info["wind_shear"].append(properties["Wind_Shear"])
        convective_init_info["low_rh"].append(properties["Low_RH"])
        convective_init_info["mid_rh"].append(properties["Mid_RH"])
        convective_init_info["richardson"].append(properties["Richardson"])
        convective_init_info["elr_0_3"].append(properties["ELR_0-3km"])

    # Create sonde data output Dataset
    sonde_output_indices_data = xr.Dataset(
        coords={"cell_id": convective_init_info["cell_id"]},
        data_vars={
            "sonde_time": ("cell_id", convective_init_info["sonde_time"]),
            "cape": ("cell_id", convective_init_info["cape"]),
            "ncape": ("cell_id", convective_init_info["ncape"]),
            "cin": ("cell_id", convective_init_info["cin"]),
            "lnb": ("cell_id", convective_init_info["lnb"]),
            "lfc": ("cell_id", convective_init_info["lfc"]),
            "lcl": ("cell_id", convective_init_info["lcl"]),
            "wind_shear": ("cell_id", convective_init_info["wind_shear"]),
            "low_rh": ("cell_id", convective_init_info["low_rh"]),
            "mid_rh": ("cell_id", convective_init_info["mid_rh"]),
            "richardson": ("cell_id", convective_init_info["richardson"]),
            "elr_0_3": ("cell_id", convective_init_info["elr_0_3"]),
        },
        attrs={
            "description": "Convective properties calculated at the time before the initation of cell formations"
        },
    )

    return sonde_output_indices_data



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jun  5 17:26:17 2024

@author: thahn
"""

# =============================================================================
# This defines the methods for running tobac on WRF data processed using wrf_load.py
# =============================================================================

import logging
from copy import deepcopy

import geopandas as gpd
import iris.cube
import numpy as np
import tobac
import xarray as xr


# Calculate nearest item in list to given pivot
def find_nearest(array, pivot):
    array = np.asarray(array)
    idx = (np.abs(array - pivot)).argmin()
    return idx


def wrf_tobac_feature_id(cube: iris.cube.Cube, CONFIG: dict) -> gpd.GeoDataFrame | None:
    """


    Parameters
    ----------
    cube : iris.cube.Cube
        Iris cube containing the variable to be tracked.
    CONFIG : dict
        User configuration file.

    Raises
    ------
    Exception
        Exception if out-of-bounds height.

    Returns
    -------
    wrf_geopd : geopandas.geodataframe.GeoDataFrame
        Geodataframe containing all default tobac feature id outputs.

    """

    feat_cube = deepcopy(cube)
    inCONFIG = deepcopy(CONFIG)

    if "height" in inCONFIG["wrf"]["tobac"]["feature_id"]:
        # Ensure segmentation_height is a proper number before running
        if (
            inCONFIG["wrf"]["tobac"]["feature_id"]["height"] is None
            or type(inCONFIG["wrf"]["tobac"]["feature_id"]["height"]) == str
            or type(CONFIG["wrf"]["tobac"]["feature_id"]["height"]) == bool
        ):
            raise Exception(
                f"""!=====Invalid Feature Identification Height. You Entered: {inCONFIG["wrf"]["tobac"]["feature_id"]["height"]}=====!"""
            )
        if (
            inCONFIG["wrf"]["tobac"]["feature_id"]["height"] * 1000
            > cube.coord("altitude").points.max()
            or inCONFIG["wrf"]["tobac"]["feature_id"]["height"] * 1000
            < cube.coord("altitude").points.min()
        ):
            raise Exception(
                f"""!=====Feature Identification Height Out of Bounds. You Entered: {inCONFIG["wrf"]["tobac"]["feature_id"]["height"]}=====!"""
            )

        # Find the nearest model height to the entered segmentation height--bypasses precision issues and allows for selection of rounded heights
        height_index = find_nearest(
            cube.coord("altitude").points,
            inCONFIG["wrf"]["tobac"]["feature_id"]["height"] * 1000,
        )

        feat_cube = feat_cube[:, height_index]
        feat_cube.remove_coord("altitude")
        feat_cube.remove_coord("model_level_number")

        del inCONFIG["wrf"]["tobac"]["feature_id"]["height"]

    # Get horozontal spacings
    dxy = tobac.get_spacings(cube)[0]

    # Perform tobac feature identification and then convert to a geodataframe before returning
    wrf_radar_features = tobac.feature_detection.feature_detection_multithreshold(
        feat_cube, dxy=dxy, **inCONFIG["wrf"]["tobac"]["feature_id"]
    )

    if wrf_radar_features is None:
        return None

    wrf_geopd = gpd.GeoDataFrame(
        wrf_radar_features,
        geometry=gpd.points_from_xy(
            wrf_radar_features.longitude, wrf_radar_features.latitude
        ),
        crs="EPSG:4326",
    )

    return wrf_geopd


def wrf_tobac_linking(
    cube: iris.cube.Cube, radar_features: gpd.GeoDataFrame, CONFIG: dict
) -> gpd.GeoDataFrame | None:
    """


    Parameters
    ----------
    cube : iris.cube.Cube
        Iris cube containing the variable to be tracked.
    radar_features : gpd.GeoDataFrame
        Tobac radar features from wrf_tobac_feature_id output.
    CONFIG : dict
        User configuration file.

    Returns
    -------
    wrf_geopd_tracks : geopandas.geodataframe.GeoDataFrame
        Geodataframe containing all default tobac tracking outputs.

    """

    if radar_features is None:
        return None

    # Mute tobac logging output
    logging.getLogger("trackpy").setLevel(level=logging.ERROR)

    dxy, dt = tobac.get_spacings(cube)

    # Do tracking then convert output dataframe to a geodataframe
    wrf_tracks = tobac.linking_trackpy(
        radar_features,
        cube,
        dt=dt,
        dxy=dxy,
        vertical_coord="altitude",
        **CONFIG["wrf"]["tobac"]["linking"],
    )

    if wrf_tracks is None:
        return None

    wrf_geopd_tracks = gpd.GeoDataFrame(
        wrf_tracks,
        geometry=gpd.points_from_xy(wrf_tracks.longitude, wrf_tracks.latitude),
        crs="EPSG:4326",
    )

    return wrf_geopd_tracks


def wrf_tobac_segmentation(
    cube: iris.cube.Cube,
    radar_features: gpd.GeoDataFrame,
    segmentation_type: str,
    CONFIG: dict,
    segmentation_height: float | None = None,
) -> tuple[xr.DataArray, gpd.GeoDataFrame]:
    """


    Parameters
    ----------
    cube : iris.cube.Cube
        Iris cube containing the variable to be tracked.
    radar_features : gpd.GeoDataFrame
        tobac radar features from wrf_tobac_feature_id output.
    segmentation_type : str
        ["2D", "3D"], whether to perform 2d segmentation or 3d segmentation.
    CONFIG : dict
        User configuration file.
    segmentation_height : float | None, optional
        height, in meters, to perform the updraft or reflectivity segmentation if 2d selected and tracking_var is not 2D. The default is None.

    Raises
    ------
    Exception
        Exception if out-of-bounds height.

    Returns
    -------
    segment_array : xarray.core.dataarray.DataArray
        Xarray DataArray containing segmented data.
    segment_features : geopandas.geodataframe.GeoDataFrame
        Geodataframe with ncells column.

    """

    if radar_features is None:
        return (None, None)

    # Enforce 2D tracking only for 2D variables
    if (len(cube.shape) == 3) and not segmentation_type.lower() == "2d":
        raise Exception(
            f"!=====Invalid Segmentation Type. You Entered: {segmentation_type.lower()}. TB and PR Tracking Restricted to 2D Segmentation=====!"
        )

    dxy = tobac.get_spacings(cube)[0]
    inCONFIG = deepcopy(CONFIG)

    # 2D and 3D segmentation have different requirements so they are split up here
    if segmentation_type.lower() == "2d":
        if "height" in inCONFIG["wrf"]["tobac"]["segmentation_2d"]:
            del inCONFIG["wrf"]["tobac"]["segmentation_2d"]["height"]

        # If altitude and/or model level number is present, remove it

        # If tracking var is 2d, bypass height
        if len(cube.shape) == 3:
            # Perform the 2d segmentation at the height_index and return the segmented cube and new geodataframe
            segment_cube, segment_features = tobac.segmentation_2D(
                radar_features,
                cube,
                dxy=dxy,
                **inCONFIG["wrf"]["tobac"]["segmentation_2d"],
            )

            # Convert iris cube to xarray and return
            # Add projection x and y back to xarray DataArray
            outXarray = xr.DataArray.from_iris(segment_cube).assign_coords(
                projection_x_coordinate=(
                    "west_east",
                    segment_cube.coord("projection_x_coordinate").points,
                ),
                projection_y_coordinate=(
                    "south_north",
                    segment_cube.coord("projection_y_coordinate").points,
                ),
            )

            return (outXarray, segment_features)

        # Ensure segmentation_height is a proper number before running
        if (
            segmentation_height == None
            or type(segmentation_height) == str
            or type(segmentation_height) == bool
        ):
            raise Exception(
                f"!=====Segmentation Height Out of Bounds. You Entered: {segmentation_height}=====!"
            )
        if (
            segmentation_height * 1000 > cube.coord("altitude").points.max()
            or segmentation_height * 1000 < cube.coord("altitude").points.min()
        ):
            raise Exception(
                f"!=====Segmentation Height Out of Bounds. You Entered: {segmentation_height}=====!"
            )

        # Find the nearest model height to the entered segmentation height--bypasses precision issues and allows for selection of rounded heights
        height_index = find_nearest(cube.coord("altitude").points, segmentation_height)

        # Remove 1 dimensional coordinates cause by taking only one altitude
        seg_cube = deepcopy(cube[:, height_index])
        seg_cube.remove_coord("altitude")
        seg_cube.remove_coord("model_level_number")

        # Perform the 2d segmentation at the height_index and return the segmented cube and new geodataframe
        segment_cube, segment_features = tobac.segmentation_2D(
            radar_features,
            seg_cube,
            dxy=dxy,
            **inCONFIG["wrf"]["tobac"]["segmentation_2d"],
        )

        # Convert iris cube to xarray and return
        # Add projection x and y back to xarray DataArray
        outXarray = xr.DataArray.from_iris(segment_cube).assign_coords(
            projection_x_coordinate=(
                "west_east",
                segment_cube.coord("projection_x_coordinate").points,
            ),
            projection_y_coordinate=(
                "south_north",
                segment_cube.coord("projection_y_coordinate").points,
            ),
        )

        return (outXarray, segment_features)

    if segmentation_type.lower() == "3d":
        # Similarly, perform 3d segmentation then return products
        segment_cube, segment_features = tobac.segmentation_3D(
            radar_features, cube, dxy=dxy, **inCONFIG["wrf"]["tobac"]["segmentation_3d"]
        )

        # Convert iris cube to xarray and return
        # Add projection x and y back to xarray DataArray
        outXarray = xr.DataArray.from_iris(segment_cube).assign_coords(
            projection_x_coordinate=(
                "west_east",
                segment_cube.coord("projection_x_coordinate").points,
            ),
            projection_y_coordinate=(
                "south_north",
                segment_cube.coord("projection_y_coordinate").points,
            ),
        )

        return (outXarray, segment_features)

    raise Exception(
        f"!=====Invalid Segmentation Type. You Entered: {segmentation_type.lower()}=====!"
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jul 10 12:29:56 2024

@author: thahn
"""

# =============================================================================
# This defines the methods for running MOAAP on WRF data processed using wrf_load.py
# =============================================================================


import os

import numpy as np
import pandas as pd
import xarray as xr

from CoCoMET.MOAAP import moaap

from .wrf_calculate_products import (
    wrf_calculate_brightness_temp,
    wrf_calculate_precip_rate,
)


# Calculate nearest item in list to given pivot
def find_nearest(array, pivot):
    array = np.asarray(array)
    idx = (np.abs(array - pivot)).argmin()
    return idx


def wrf_run_moaap(wrf_xarray: xr.Dataset, CONFIG: dict) -> xr.Dataset:
    """


    Parameters
    ----------
    wrf_xarray : xarray.core.dataset.Dataset
        Xarray Dataset containing WRF data calculated from wrf_load.py.
    CONFIG : dict
        User configuration file.

    Returns
    -------
    mask_file : xarray.core.dataset.Dataset
        The xarray object containing the default MOAAP outputs.

    """

    # Get basic setup variables including lat/lon, delta time, a pandas time range vector (TODO: adjust output to )
    latitudes = wrf_xarray.XLAT[0].values
    longitudes = wrf_xarray.XLONG[0].values
    dt = wrf_xarray.DT  # in s
    times = pd.date_range(
        start=wrf_xarray.XTIME[0].values,  # .astype('datetime64[s]'),
        end=wrf_xarray.XTIME[-1].values,  # .astype('datetime64[s]'),
        freq=str(dt / 60) + "min",
    )
    mask = np.ones(latitudes.shape)

    # Get all necessary variables from WRF output to input into MOAAP

    # Get pressure heights
    p = wrf_xarray["P"]
    pb = wrf_xarray["PB"]

    total_p = (p + pb) / 100
    avg_geo_pres = [np.mean(h.values) for h in total_p[0]]

    # Get height_idx of 850hPA, 500hPA, and 200hPA
    height_idx_850 = find_nearest(avg_geo_pres, 850)
    height_idx_500 = find_nearest(avg_geo_pres, 500)
    height_idx_200 = find_nearest(avg_geo_pres, 200)

    # get destaggered 850hPA wind speeds and 200hPA wind speeds
    v_winds_850 = (
        0.5 * wrf_xarray["V"][:, height_idx_850, 1:, :]
        + 0.5 * wrf_xarray["V"][:, height_idx_850, :-1, :]
    ).values
    u_winds_850 = (
        0.5 * wrf_xarray["U"][:, height_idx_850, :, 1:]
        + 0.5 * wrf_xarray["U"][:, height_idx_850, :, :-1]
    ).values
    v_winds_200 = (
        0.5 * wrf_xarray["V"][:, height_idx_200, 1:, :]
        + 0.5 * wrf_xarray["V"][:, height_idx_200, :-1, :]
    ).values
    u_winds_200 = (
        0.5 * wrf_xarray["U"][:, height_idx_200, :, 1:]
        + 0.5 * wrf_xarray["U"][:, height_idx_200, :, :-1]
    ).values

    # Get 850hPA air temperature
    t = wrf_xarray["T"]

    # Calculate proper pressures and actual temperature
    full_t = t + 300
    full_p = p + pb

    air_temp = (full_t * (full_p / 1e5) ** (287.0 / 1004.5))[:, height_idx_850].values

    # Get geopotential heights
    ph = wrf_xarray["PH"]
    phb = wrf_xarray["PHB"]
    geopt = ph + phb

    # DESTAGGER geopt
    geopt = 0.5 * geopt[:, 1:] + 0.5 * geopt[:, :-1]

    # Get brightness temp
    tb = (
        wrf_xarray.TB.values
        if "TB" in wrf_xarray
        else wrf_calculate_brightness_temp(wrf_xarray)
    )

    # Get total mixing ratio at 850hPA
    mr = (
        wrf_xarray.QVAPOR
        + wrf_xarray.QCLOUD
        + wrf_xarray.QRAIN
        + wrf_xarray.QICE
        + wrf_xarray.QSNOW
        + wrf_xarray.QGRAUP
    )[:, height_idx_850]

    # Get precipitation rate
    pr = wrf_calculate_precip_rate(wrf_xarray) / (3600 / wrf_xarray.DT)

    # Create output folder if it does not already exist
    if not os.path.exists(CONFIG["wrf"]["moaap"]["tracking_save_path"]):
        os.makedirs(CONFIG["wrf"]["moaap"]["tracking_save_path"])

    moaap(
        longitudes,
        latitudes,
        times,
        dt / 3600,  # convert to hr
        mask,
        DataName="CoMET_WRF_MOAAP_TRACKING",
        OutputFolder=CONFIG["wrf"]["moaap"]["tracking_save_path"],
        # Data Variables
        v850=v_winds_850,
        u850=u_winds_850,
        t850=air_temp,
        q850=mr,
        slp=full_p[:, 0],
        ivte=None,
        ivtn=None,
        z500=geopt.values[:, height_idx_500],
        v200=v_winds_200,
        u200=u_winds_200,
        pr=pr,
        tb=tb,
        # Any user defined params
        **CONFIG["wrf"]["moaap"],
    )

    output_filepath = (
        CONFIG["wrf"]["moaap"]["tracking_save_path"]
        + str(times[0].year)
        + str(times[0].month).zfill(2)
        + "_CoMET_WRF_MOAAP_TRACKING_ObjectMasks_dt-%.2f" % (dt / 60)
        + "min_MOAAP-masks"
        + ".nc"
    )

    mask_file = xr.open_mfdataset(
        output_filepath,
        coords="all",
        concat_dim="time",
        combine="nested",
        decode_times=False,
    )

    return mask_file



"""
This file contains functions for dealing with the dynamic vertical scaling when working with 3d data
Some functions in here rely on others, so its best to import the entire file
"""


# TODO: This file is a disaster, needs to be heavily rewritten
def calc_3d_perim(footprint_data, frame, feature_id):
    """
    Inputs:
        footprint_data: US_segmentation_3d.Feature_Segmentation
        frame: frame number of the relevant cell
        feature_id: feature_id number of the relevant cell
    Outputs:
        perims: A dictionary with keys equal to what the surface area is touching (-1 for empty space, feature_id of other cells) and values equal to the amount of surface area shared with that object
    """

    from copy import deepcopy

    import numpy as np
    from dask.array import isin

    final_mask = deepcopy(footprint_data[frame, :, :, :])
    feature_mask = deepcopy(final_mask)
    feature_mask.values[~isin(final_mask.values, feature_id)] = -1
    # tilde inverts the output, so this creates a boolean array with True where feature_id is, then inverts that to
    # set all non-feature_id values to -1

    # If cell has no area, skip
    if np.all(feature_mask.values == -1):
        return -1

    # see if this cell is touching any other cells
    feature_data = np.asarray(final_mask.values)

    z_dim_sizes = np.diff(footprint_data.altitude)
    x_dim_sizes = np.diff(footprint_data.projection_x_coordinate)
    y_dim_sizes = np.diff(footprint_data.projection_y_coordinate)

    # These are one cell too small due to how diff works, so infer last cell size using the same cell size as the previous cell
    x_dim_sizes = np.append(x_dim_sizes, x_dim_sizes[-1])
    y_dim_sizes = np.append(y_dim_sizes, y_dim_sizes[-1])
    z_dim_sizes = np.append(z_dim_sizes, z_dim_sizes[-1])

    x_dim_sizes = x_dim_sizes / 1000  # convert to km
    y_dim_sizes = y_dim_sizes / 1000
    z_dim_sizes = z_dim_sizes / 1000

    perims = (
        {}
    )  # this is to keep track of the amount of the perimeter that is shared with another cell
    # key is feature id of other cell, value is the shared perimeter

    for nz, ny, nx in list(zip(*np.where(feature_data == feature_id))):
        # print(nx, ny, nz)

        # Find if location is on edge of cell (i.e. any touching cells not)
        neighboring_cells = []

        for mx in (nx - 1, nx + 1):
            if mx in range(feature_data.shape[2]):
                t = feature_data[nz, ny, mx]
            else:
                t = -1
            if t != feature_id:
                perims[t] = perims.get(t, 0) + y_dim_sizes[ny] * z_dim_sizes[nz]

        for my in (ny - 1, ny + 1):
            if my in range(feature_data.shape[1]):
                t = feature_data[nz, my, nx]
            else:
                t = -1
            if t != feature_id:
                perims[t] = perims.get(t, 0) + x_dim_sizes[nx] * z_dim_sizes[nz]

        for mz in (nz - 1, nz + 1):
            if mz in range(feature_data.shape[0]):
                t = feature_data[mz, ny, nx]
            else:
                t = -1
            if t != feature_id:
                perims[t] = perims.get(t, 0) + x_dim_sizes[nx] * y_dim_sizes[ny]
    return perims


def calculate_volume(mask, points):
    """
    Inputs:
        mask: US_segmentation_3d.Feature_Segmentation
        points: list of (z, y, x) grid points that make up the cell
    Outputs:
        volume: volume of cell in km^3
    """

    import numpy as np

    # We first calculate the area of each individual cell
    # First get the size of each dimension
    x_dim_sizes = np.diff(mask.projection_x_coordinate)
    y_dim_sizes = np.diff(mask.projection_y_coordinate)
    z_dim_sizes = np.diff(mask.altitude)

    # These are one cell too small due to how diff works, so infer last cell size using the same cell size as the previous cell
    # x_dim_sizes.append(x_dim_sizes[-1])
    x_dim_sizes = np.append(x_dim_sizes, x_dim_sizes[-1])
    y_dim_sizes = np.append(y_dim_sizes, y_dim_sizes[-1])
    z_dim_sizes = np.append(z_dim_sizes, z_dim_sizes[-1])

    # Why aren't we using this?
    # use Einstein sum notation to get volume of cells
    # cell_volumes = np.einsum("i,j,k->ijk", z_dim_sizes, y_dim_sizes, x_dim_sizes)

    total_vol = 0
    for z, y, x in points:
        total_vol += x_dim_sizes[x] * y_dim_sizes[y] * z_dim_sizes[z]

    return total_vol / (10**9)


def create_gridded_bounds(footprint_data, center, radius):
    """
    Inputs:
        footprint_data: US_Segmentatio_3d. Must have altitude, x projection, and y projection as attributes
        center: center of the cube (usually feature position given by Tracks/Feature df) in projection coordinates (z, y, x)
        radius: half of side length of the desired cube in km
    Outputs:
        tuple of the minimum and maximum grid points of z, y, x as tuples (min,max)
    """

    altitudes = (
        footprint_data.altitude
    )  # get the projection values at every grid coordinate point
    yprojs = footprint_data.projection_y_coordinate
    xprojs = footprint_data.projection_x_coordinate

    altitude = center[0]
    proj_y = center[1]
    proj_x = center[2]

    max_alt = altitude + (
        radius * 1000
    )  # get the bounds of the box in terms of projection coords
    min_alt = altitude - (radius * 1000)
    max_y_proj = proj_y + (radius * 1000)
    min_y_proj = proj_y - (radius * 1000)
    max_x_proj = proj_x + (radius * 1000)
    min_x_proj = proj_x - (radius * 1000)

    # Convert to grid points
    max_z = find_nearest(altitudes, max_alt)  # convert bounds back to grid coords
    min_z = find_nearest(altitudes, min_alt)
    max_y = find_nearest(yprojs, max_y_proj)
    min_y = find_nearest(yprojs, min_y_proj)
    max_x = find_nearest(xprojs, max_x_proj)
    min_x = find_nearest(xprojs, min_x_proj)

    return ((min_z, max_z), (min_y, max_y), (min_x, max_x))


def find_nearest(array, pivot):
    # Calculate nearest item in list to given pivot
    # This is used when converting projection coordinates into grid coordinate indices
    import numpy as np

    array = np.asarray(array)
    idx = (np.abs(array - pivot)).argmin()
    return idx


def create_scaled_3d_mesh(
    footprint_data, frame, feature_id, variable_field=None, step_size=1
):
    """
    Inputs:
        footprint_data: Xarray. US_segmentation_3d.Feature_Segmentation
        variable_field: Xarray. Xarray of variable that was used for segmentation
        frame: Int. frame of the specified cell
        feature_id: Int. feature_id of the specified cell
        step_size: Int. Default is 1. Step size input to marching cubes algorithm. Larger number means coarser mesh.
    Outputs:
        verts: (V, 3) Array. Vertices of the mesh in projection coordinates (altitude, projection_y, projection_x) meters
        faces: (F, 3) Array. List of tuples describing the faces of the mesh. Each entry of the tuple is the index of a vertex in verts, and 3 indices in a tuple make up one face.
        values: (V,) Array. Maximum value of the data in the local region near each vertex. Usually used for visualization
    """
    from copy import deepcopy

    import numpy as np
    from skimage.measure import marching_cubes

    final_mask = deepcopy(footprint_data[frame])
    data = final_mask.values
    mask = data == feature_id
    thresh = 0.5
    if type(variable_field) != type(None):
        mask = (
            variable_field[frame, :, :, :].where(mask, -1).values
        )  # create a mask with the variable value where the segmentation is and -1 everywhere else
        thresh = 0

    verts, faces, normals, values = marching_cubes(mask, thresh, step_size=step_size)

    verts_proj = []
    for vertex in verts:  # convert each vertex to projection coordinates
        vertex_projection = point_projection(footprint_data, vertex)
        verts_proj.append(vertex_projection / 1000)
    verts_proj = np.array(verts_proj)
    return verts_proj, faces, values


def point_projection(footprint_data, point):
    """
    Inputs:
        footprint_data: Xarray. US_segmentation_3d.Feature_Segmentation
        point: Array-like. 3d point to be converted from grid coordinates to projection coordinates in form (z, y, x),
        Or 2d point in (x, y)
    Outputs:
        projection_point: List. Projection coordinates of input point in form [z, y, x]
    """

    import numpy as np

    projection_x = (
        footprint_data.projection_x_coordinate.values
    )  # get projection values for every grid coordinate
    projection_y = footprint_data.projection_y_coordinate.values
    altitudes = footprint_data.altitude.values

    if len(point) == 2:
        x = point[0]
        y = point[1]
        coord_list = [[x, projection_x], [y, projection_y]]
    elif len(point) == 3:
        x = point[2]
        y = point[1]
        z = point[0]
        coord_list = [[z, altitudes], [y, projection_y], [x, projection_x]]
    else:
        raise ValueError("Point should either be 2 or 3 dimensional")

    projection_point = []  # output point [z,y,x] or [x,y]

    for i, projection in coord_list:
        # for each coord z,y,x interpolate linearly between the projection bounds to find the projection value at the point
        upper_bound = projection[int(np.ceil(i))]
        lower_bound = projection[int(np.floor(i))]
        difference = upper_bound - lower_bound

        fraction = i - np.trunc(
            i
        )  # fractional component of the input coord (i.e. 1.4 -> 0.4)

        fract_diff = fraction * difference

        projection_i = lower_bound + fract_diff
        projection_point.append(projection_i)

    return np.array(projection_point)


def distance_projection(footprint_data, point1, point2):
    import numpy as np

    proj1 = point_projection(footprint_data, point1)
    proj2 = point_projection(footprint_data, point2)

    z1, y1, x1 = proj1
    z2, y2, x2 = proj2

    dist = np.sqrt((z2 - z1) ** 2 + (y2 - y1) ** 2 + (x2 - x1) ** 2)

    return dist



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Jul 15 11:57:43 2024

@author: thahn
"""

# =============================================================================
# This file contains scripts for loading (and gridding) multiple NEXRAD radars at once
# =============================================================================

import glob
import multiprocessing
import warnings
from functools import partial

import cftime
import iris
import iris.cube
import numpy as np
import pyart
import xarray as xr
from tqdm import tqdm


# Calculate nearest item in list to given pivot
def find_nearest(array, pivot):
    array = np.asarray(array)
    idx = (np.abs(array - pivot)).argmin()
    return idx


def gen_and_save_multi_nexrad_grid(
    paths_to_files: list[str],
    save_location: str,
    tracking_var: str,
    CONFIG: dict,
    parallel_processing: bool = False,
    max_cores: int | None = None,
) -> None:
    """


    Parameters
    ----------
    paths_to_files : list[str]
        Array of glob path to archival NEXRAD level 2 input files--i.e. ["/data/usr/KVNX*_V06.ar2v", "/data/usr/KIVX*_V06.ar2v"].
    save_location : str
        Path to where the gridded NEXRAD files should be saved to, should be a directory and end with "/".
    tracking_var : str
        ["dbz"], variable which is going to be used for tracking--reflectivity.
    CONFIG : dict
        User configuration file.
    parallel_processing : bool, optional
        Bool determinig whether to use parallel processing when gridding files. The default is False.
    max_cores : int, optional
        Number of cores to use if parallel_processing == True. The default is None.

    Raises
    ------
    Exception
        Exception if there are no files to grid or invalid tracking variable entered.

    Returns
    -------
    None

    """

    # No parallel processing for this as the time of arrival matters
    radar_list = []
    radar_time_list = []

    for ii, radar_path in enumerate(paths_to_files):
        radar_list.append([])
        radar_time_list.append([])

        # Get all archive files and iterate over them
        files = np.sort(glob.glob(radar_path))

        if len(files) == 0:
            raise Exception("!=====No Files Present to Grid=====!")

        # Get radars and their respective times. Don"t load them all at once to conserve memory
        for ff in tqdm(
            files, desc=f"=====Loading NEXRAD #{ii+1}=====", total=len(files)
        ):
            if tracking_var.lower() == "dbz":
                # Create radar object including only field of interest
                radar = pyart.io.read_nexrad_archive(ff, include_fields="reflectivity")

                radar_list[ii].append(ff)

                # Add datetime of radar
                radar_time_list[ii].append(
                    pyart.util.datetime_from_radar(
                        radar,
                    )
                )

                del radar

            else:
                raise Exception(
                    f"!=====Invalid Tracking Variable. You Entered: {tracking_var.lower()}=====!"
                )

    # First get all lenghts
    radar_lens = np.array([len(f) for f in radar_list])

    # If parallel processing, save grids multi
    if parallel_processing:
        # Start a pool with max_cores and run the grid function
        with multiprocessing.Pool(max_cores) as multi_pool:
            with tqdm(
                total=np.min(radar_lens),
                desc="=====Parallel Gridding Combined NEXRAD=====",
            ) as pbar:
                for _ in multi_pool.imap_unordered(
                    partial(
                        parallel_save_grid,
                        save_location=save_location,
                        radar_list=radar_list,
                        radar_time_list=radar_time_list,
                        CONFIG=CONFIG,
                    ),
                    np.arange(0, np.min(radar_lens)),
                ):
                    pbar.update()

        # Delete remaining radar arrays
        del radar_list
        del radar_time_list

        return

    # Now loop over shortest subset of radar objects

    # Find the shortest radar and get the index in the radar list of said radar
    shortest_radar_idx = np.where(radar_lens == np.min(radar_lens))[0][0]

    # Do this to make the for loop indexing slightly easier
    shortest_radar_list = radar_list[shortest_radar_idx]
    shortest_radar_time_list = radar_time_list[shortest_radar_idx]

    radar_list.pop(shortest_radar_idx)
    radar_time_list.pop(shortest_radar_idx)

    # Loop over shortest length
    for ii in tqdm(
        range(np.min(radar_lens)),
        desc="=====Gridding Combined NEXRAD Data=====",
        total=np.min(radar_lens),
    ):
        # Keep track of the radars we are merging and the average position of them
        merge_radar_list = []
        lats = []
        lons = []

        # Append the base radar
        base_radar = pyart.io.read_nexrad_archive(
            shortest_radar_list[ii], include_fields="reflectivity"
        )
        merge_radar_list.append(base_radar)
        lats.append(base_radar.latitude["data"][0])
        lons.append(base_radar.longitude["data"][0])

        # For each additional radar, add the nearest time and then its position
        for jj, additional_radar in enumerate(radar_list):
            # Grab the nearest radar object to the shortest iith one for every other radar
            nearest_time_idx = find_nearest(
                radar_time_list[jj], shortest_radar_time_list[ii]
            )

            addit_radar = pyart.io.read_nexrad_archive(
                additional_radar[nearest_time_idx], include_fields="reflectivity"
            )
            merge_radar_list.append(addit_radar)

            lats.append(addit_radar.latitude["data"][0])
            lons.append(addit_radar.longitude["data"][0])

            del addit_radar

        # Now merge the radars
        combined_grid = pyart.map.grid_from_radars(
            merge_radar_list,
            grid_origin=(np.mean(lats), np.mean(lons)),
            **CONFIG["multi_nexrad"]["gridding"],
        )

        # Save radar grid to save_location as a netcdf file
        shortest_radar_time_ii = shortest_radar_time_list[ii].strftime(
            "%Y_%m_%d_%H:%M:%S"
        )
        pyart.io.write_grid(
            save_location + f"combined_grid_{shortest_radar_time_ii}.nc",
            combined_grid,
            arm_alt_lat_lon_variables=True,
            write_point_x_y_z=True,
            write_point_lon_lat_alt=True,
        )

        del base_radar
        del combined_grid
        del merge_radar_list

    # Delete remaining radar arrays
    del radar_list
    del radar_time_list

    return


def parallel_save_grid(ii, save_location, radar_list, radar_time_list, CONFIG):
    """
    Helper function for multi-processing
    """

    # Now loop over shortest subset of radar objects
    radar_lens = np.array([len(f) for f in radar_list])

    # Find the shortest radar and get the index in the radar list of said radar
    shortest_radar_idx = np.where(radar_lens == np.min(radar_lens))[0][0]

    # Do this to make the for loop indexing slightly easier
    shortest_radar_list = radar_list[shortest_radar_idx]
    shortest_radar_time_list = radar_time_list[shortest_radar_idx]

    radar_list.pop(shortest_radar_idx)
    radar_time_list.pop(shortest_radar_idx)

    # Keep track of the radars we are merging and the average position of them
    merge_radar_list = []
    lats = []
    lons = []

    # Append the base radar
    base_radar = pyart.io.read_nexrad_archive(
        shortest_radar_list[ii], include_fields="reflectivity"
    )
    merge_radar_list.append(base_radar)
    lats.append(base_radar.latitude["data"][0])
    lons.append(base_radar.longitude["data"][0])

    # For each additional radar, add the nearest time and then its position
    for jj, additional_radar in enumerate(radar_list):
        # Grab the nearest radar object to the shortest iith one for every other radar
        nearest_time_idx = find_nearest(
            radar_time_list[jj], shortest_radar_time_list[ii]
        )

        addit_radar = pyart.io.read_nexrad_archive(
            additional_radar[nearest_time_idx], include_fields="reflectivity"
        )
        merge_radar_list.append(addit_radar)

        lats.append(addit_radar.latitude["data"][0])
        lons.append(addit_radar.longitude["data"][0])

        del addit_radar

    # Now merge the radars
    combined_grid = pyart.map.grid_from_radars(
        merge_radar_list,
        grid_origin=(np.mean(lats), np.mean(lons)),
        **CONFIG["multi_nexrad"]["gridding"],
    )

    # Save radar grid to save_location as a netcdf file
    pyart.io.write_grid(
        save_location
        + f"""combined_grid_{shortest_radar_time_list[ii].strftime("%Y_%m_%d_%H:%M:%S")}.nc""",
        combined_grid,
        arm_alt_lat_lon_variables=True,
        write_point_x_y_z=True,
        write_point_lon_lat_alt=True,
    )

    del base_radar
    del combined_grid
    del merge_radar_list

    return


def multi_nexrad_load_netcdf_iris(
    paths_to_files: list[str] | str,
    file_type: str,
    tracking_var: str,
    CONFIG: dict,
    save_location: str | None = None,
) -> tuple[iris.cube.Cube, xr.DataArray]:
    """


    Parameters
    ----------
    paths_to_files : list[str] | str
        Array of glob path to input files, either archival or grided netcdf--i.e. ["/data/usr/KVNX*_V06.ar2v", "/data/usr/KIVX*_V06.ar2v"]. ONLY AN ARRAY WHEN NOT GRIDDED YET.
    file_type : str
        ["ar2v", "nc"] type of input file--either archival or netcdf.
    tracking_var : str
        ["dbz"], variable which is going to be used for tracking--reflectivity.
    CONFIG : dict
        User configuration file.
    save_location : str, optional
        Where to save gridded NEXRAD data to if file_type=="ar2v". The default is None.

    Raises
    ------
    Exception
        Exception if missing MULTI-NEXRAD field in CONFIG, invalid tracking variable, and/or invalid file type.

    Returns
    -------
    nexrad_cube : iris.cube.Cube
        Iris cube continaing gridded reflectivity data ready for tobac tracking.
    nexrad_xarray : xarray.core.dataarray.DataArray
        Xarray DataArray containing gridded NEXRAD archival data.

    """

    # If data is archival, perform gridding
    if file_type.lower() == "ar2v":
        # Make sure save_location ends with a "/"
        if save_location[-1] != "/":
            save_location = save_location + "//"

        # Create grid
        gen_and_save_multi_nexrad_grid(
            paths_to_files,
            save_location,
            tracking_var,
            CONFIG,
            CONFIG["parallel_processing"],
            CONFIG["max_cores"],
        )

        if tracking_var.lower() == "dbz":
            # Read radar objects in and concat into one xarray datarray
            # This is a stupid hacky fix because pyart is dumb
            radar_objects = []
            for file in np.sort(glob.glob(save_location + "*")):
                # Ignore import warnings
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore")
                    radar_objects.append(pyart.io.read_grid(file).to_xarray())

            nexrad_xarray = xr.concat(radar_objects, dim="time").reflectivity.drop_vars(
                [
                    "origin_altitude",
                    "origin_latitude",
                    "origin_longitude",
                    "ProjectionCoordinateSystem",
                    "projection",
                ]
            )
            del radar_objects

            # Subset location and time of interest
            if "multi_nexrad" in CONFIG:
                # Subset time based on user inputs
                if (
                    "min_frame_index" in CONFIG["multi_nexrad"]
                    or "max_frame_index" in CONFIG["multi_nexrad"]
                ):
                    min_frame = (
                        CONFIG["multi_nexrad"]["min_frame_index"]
                        if "min_frame_index" in CONFIG["multi_nexrad"]
                        else 0
                    )
                    max_frame = (
                        CONFIG["multi_nexrad"]["max_frame_index"] + 1
                        if "max_frame_index" in CONFIG["multi_nexrad"]
                        else nexrad_xarray.dims["time"]
                    )

                    nexrad_xarray = nexrad_xarray.isel(
                        time=np.arange(
                            min_frame,
                            max_frame,
                        ),
                        drop=True,
                    )

                if "bounds" in CONFIG["multi_nexrad"]:
                    mask_lon = (
                        nexrad_xarray.lon >= CONFIG["multi_nexrad"]["bounds"][0]
                    ) & (nexrad_xarray.lon <= CONFIG["multi_nexrad"]["bounds"][1])
                    mask_lat = (
                        nexrad_xarray.lat >= CONFIG["multi_nexrad"]["bounds"][2]
                    ) & (nexrad_xarray.lat <= CONFIG["multi_nexrad"]["bounds"][3])

                    nexrad_xarray = nexrad_xarray.where(mask_lon & mask_lat, drop=True)

            else:
                raise Exception("""!=====CONFIG Missing "multi_nexrad" Field=====!""")

            # Replace time dimension with minutes since first time and add other x y z coords
            first_time = nexrad_xarray.time.values[0]
            nexrad_xarray = nexrad_xarray.assign_coords(
                time=cftime.date2num(
                    nexrad_xarray.time.values, f"minutes since {first_time}"
                ),
                south_north=("y", np.arange(nexrad_xarray.shape[2])),
                west_east=("x", np.arange(nexrad_xarray.shape[3])),
                projection_x_coordinate=("x", nexrad_xarray.x.values),
                projection_y_coordinate=("y", nexrad_xarray.y.values),
                x=("x", np.arange(nexrad_xarray.shape[3])),
                y=("y", np.arange(nexrad_xarray.shape[2])),
                model_level_number=("z", np.arange(nexrad_xarray.shape[1])),
            )

            # Create DT attribute
            dt_array = np.diff(nexrad_xarray.time.values)
            if len(np.unique(dt_array)) != 1:
                nexrad_xarray.attrs["DT"] = dt_array * 60  # min -> s
            else:
                nexrad_xarray.attrs["DT"] = dt_array[0] * 60  # min -> s

            # Adjust dimension names to be standards accepted by iris
            nexrad_xarray["time"] = nexrad_xarray.time.assign_attrs(
                {
                    "standard_name": "time",
                    "long_name": f"minutes since {first_time}",
                    "units": f"minutes since {first_time}",
                }
            )
            nexrad_xarray["z"] = nexrad_xarray.z.assign_attrs(
                {"standard_name": "altitude", "units": "m"}
            )
            nexrad_xarray["lat"] = nexrad_xarray.lat.assign_attrs(
                {"standard_name": "latitude"}
            )
            nexrad_xarray["lon"] = nexrad_xarray.lon.assign_attrs(
                {"standard_name": "longitude"}
            )
            nexrad_xarray["projection_x_coordinate"] = (
                nexrad_xarray.projection_x_coordinate.assign_attrs({"units": "m"})
            )
            nexrad_xarray["projection_y_coordinate"] = (
                nexrad_xarray.projection_y_coordinate.assign_attrs({"units": "m"})
            )

            # Add altitude dimension to xarray but not to cube
            nexrad_cube = nexrad_xarray.to_iris()
            nexrad_xarray = nexrad_xarray.assign_coords(
                altitude=("z", nexrad_xarray.z.values)
            )

            nexrad_xarray["z"] = nexrad_xarray.z.assign_attrs({"standard_name": ""})
            nexrad_xarray["altitude"] = nexrad_xarray.altitude.assign_attrs(
                {"standard_name": "altitude", "units": "m"}
            )

            return (nexrad_cube, nexrad_xarray)

        raise Exception(
            f"!=====Invalid Tracking Variable. You Entered: {tracking_var.lower()}=====!"
        )

    # If data already grided, just return concated netcdf dataset
    if file_type.lower() == "nc":
        # Convert to iris cube and return
        if tracking_var.lower() == "dbz":
            # Read radar objects in and concat into one xarray datarray
            # This is a stupid hacky fix because pyart is dumb
            radar_objects = []
            for file in np.sort(glob.glob(paths_to_files)):
                # Ignore import warnings
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore")
                    radar_objects.append(pyart.io.read_grid(file).to_xarray())

            nexrad_xarray = xr.concat(radar_objects, dim="time").reflectivity.drop_vars(
                [
                    "origin_altitude",
                    "origin_latitude",
                    "origin_longitude",
                    "ProjectionCoordinateSystem",
                    "projection",
                ]
            )
            del radar_objects

            # Subset location and time of interest
            if "multi_nexrad" in CONFIG:
                # Subset time based on user inputs
                if (
                    "min_frame_index" in CONFIG["multi_nexrad"]
                    or "max_frame_index" in CONFIG["multi_nexrad"]
                ):
                    min_frame = (
                        CONFIG["multi_nexrad"]["min_frame_index"]
                        if "min_frame_index" in CONFIG["multi_nexrad"]
                        else 0
                    )
                    max_frame = (
                        CONFIG["multi_nexrad"]["max_frame_index"] + 1
                        if "max_frame_index" in CONFIG["multi_nexrad"]
                        else nexrad_xarray.dims["time"]
                    )

                    nexrad_xarray = nexrad_xarray.isel(
                        time=np.arange(
                            min_frame,
                            max_frame,
                        ),
                        drop=True,
                    )

                if "bounds" in CONFIG["multi_nexrad"]:
                    mask_lon = (
                        nexrad_xarray.lon >= CONFIG["multi_nexrad"]["bounds"][0]
                    ) & (nexrad_xarray.lon <= CONFIG["multi_nexrad"]["bounds"][1])
                    mask_lat = (
                        nexrad_xarray.lat >= CONFIG["multi_nexrad"]["bounds"][2]
                    ) & (nexrad_xarray.lat <= CONFIG["multi_nexrad"]["bounds"][3])

                    nexrad_xarray = nexrad_xarray.where(mask_lon & mask_lat, drop=True)

            else:
                raise Exception("""!=====CONFIG Missing "multi_nexrad" Field=====!""")

            # Replace time dimension with minutes since first time and add other x y z coords
            first_time = nexrad_xarray.time.values[0]
            nexrad_xarray = nexrad_xarray.assign_coords(
                time=cftime.date2num(
                    nexrad_xarray.time.values, f"minutes since {first_time}"
                ),
                south_north=("y", np.arange(nexrad_xarray.shape[2])),
                west_east=("x", np.arange(nexrad_xarray.shape[3])),
                projection_x_coordinate=("x", nexrad_xarray.x.values),
                projection_y_coordinate=("y", nexrad_xarray.y.values),
                x=("x", np.arange(nexrad_xarray.shape[3])),
                y=("y", np.arange(nexrad_xarray.shape[2])),
                model_level_number=("z", np.arange(nexrad_xarray.shape[1])),
            )

            # Create DT attribute
            dt_array = np.diff(nexrad_xarray.time.values)
            if len(np.unique(dt_array)) != 1:
                nexrad_xarray.attrs["DT"] = dt_array * 60  # min -> s
            else:
                nexrad_xarray.attrs["DT"] = dt_array[0] * 60  # min -> s

            # Adjust dimension names to be standards accepted by iris
            nexrad_xarray["time"] = nexrad_xarray.time.assign_attrs(
                {
                    "standard_name": "time",
                    "long_name": f"minutes since {first_time}",
                    "units": f"minutes since {first_time}",
                }
            )
            nexrad_xarray["z"] = nexrad_xarray.z.assign_attrs(
                {"standard_name": "altitude", "units": "m"}
            )
            nexrad_xarray["lat"] = nexrad_xarray.lat.assign_attrs(
                {"standard_name": "latitude"}
            )
            nexrad_xarray["lon"] = nexrad_xarray.lon.assign_attrs(
                {"standard_name": "longitude"}
            )
            nexrad_xarray["projection_x_coordinate"] = (
                nexrad_xarray.projection_x_coordinate.assign_attrs({"units": "m"})
            )
            nexrad_xarray["projection_y_coordinate"] = (
                nexrad_xarray.projection_y_coordinate.assign_attrs({"units": "m"})
            )

            # Add altitude dimension to xarray but not to cube
            nexrad_cube = nexrad_xarray.to_iris()
            nexrad_xarray = nexrad_xarray.assign_coords(
                altitude=("z", nexrad_xarray.z.values)
            )

            nexrad_xarray["z"] = nexrad_xarray.z.assign_attrs({"standard_name": ""})
            nexrad_xarray["altitude"] = nexrad_xarray.altitude.assign_attrs(
                {"standard_name": "altitude", "units": "m"}
            )

            return (nexrad_cube, nexrad_xarray)

        else:
            raise Exception(
                f"!=====Invalid Tracking Variable. You Entered: {tracking_var.lower()}=====!"
            )

    raise Exception(f"!=====Invalid File Type. You Entered: {file_type.lower()}=====!")


def multi_nexrad_load_netcdf(
    paths_to_files: list[str] | str,
    file_type: str,
    tracking_var: str,
    CONFIG: dict,
    save_location: str | None = None,
) -> xr.DataArray:
    """


    Parameters
    ----------
    paths_to_files : list[str] | str
        Array of glob path to input files, either archival or grided netcdf--i.e. ["/data/usr/KVNX*_V06.ar2v", "/data/usr/KIVX*_V06.ar2v"]. ONLY AN ARRAY WHEN NOT GRIDDED YET.
    file_type : str
        ["ar2v", "nc"] type of input file--either archival or netcdf.
    tracking_var : str
        ["dbz"], variable which is going to be used for tracking--reflectivity.
    CONFIG : dict
        User configuration file.
    save_location : str, optional
        Where to save gridded NEXRAD data to if file_type=="ar2v". The default is None.

    Raises
    ------
    Exception
        Exception if missing MULTI-NEXRAD field in CONFIG, invalid tracking variable, and/or invalid file type.

    Returns
    -------
    nexrad_xarray : xarray.core.dataarray.DataArray
        Xarray DataArray containing gridded NEXRAD archival data.

    """

    # If data is archival, perform gridding
    if file_type.lower() == "ar2v":
        # Make sure save_location ends with a "/"
        if save_location[-1] != "/":
            save_location = save_location + "//"

        # Create grid
        gen_and_save_multi_nexrad_grid(
            paths_to_files,
            save_location,
            tracking_var,
            CONFIG,
            CONFIG["parallel_processing"],
            CONFIG["max_cores"],
        )

        # Open them as netcdf file and return
        if tracking_var.lower() == "dbz":
            # Read radar objects in and concat into one xarray datarray
            # This is a stupid hacky fix because pyart is dumb
            radar_objects = []
            for file in np.sort(glob.glob(save_location + "*")):
                # Ignore import warnings
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore")
                    radar_objects.append(pyart.io.read_grid(file).to_xarray())

            nexrad_xarray = xr.concat(radar_objects, dim="time").reflectivity
            del radar_objects

            # Subset location and time of interest
            if "multi_nexrad" in CONFIG:
                # Subset time based on user inputs
                if (
                    "min_frame_index" in CONFIG["multi_nexrad"]
                    or "max_frame_index" in CONFIG["multi_nexrad"]
                ):
                    min_frame = (
                        CONFIG["multi_nexrad"]["min_frame_index"]
                        if "min_frame_index" in CONFIG["multi_nexrad"]
                        else 0
                    )
                    max_frame = (
                        CONFIG["multi_nexrad"]["max_frame_index"] + 1
                        if "max_frame_index" in CONFIG["multi_nexrad"]
                        else nexrad_xarray.dims["time"]
                    )

                    nexrad_xarray = nexrad_xarray.isel(
                        time=np.arange(
                            min_frame,
                            max_frame,
                        ),
                        drop=True,
                    )

                if "bounds" in CONFIG["multi_nexrad"]:
                    mask_lon = (
                        nexrad_xarray.lon >= CONFIG["multi_nexrad"]["bounds"][0]
                    ) & (nexrad_xarray.lon <= CONFIG["multi_nexrad"]["bounds"][1])
                    mask_lat = (
                        nexrad_xarray.lat >= CONFIG["multi_nexrad"]["bounds"][2]
                    ) & (nexrad_xarray.lat <= CONFIG["multi_nexrad"]["bounds"][3])

                    nexrad_xarray = nexrad_xarray.where(mask_lon & mask_lat, drop=True)

            else:
                raise Exception("""!=====CONFIG Missing "nexrad" Field=====!""")

            # Replace time dimension with minutes since first time and add other x y z coords
            first_time = nexrad_xarray.time.values[0]
            nexrad_xarray = nexrad_xarray.assign_coords(
                time=cftime.date2num(
                    nexrad_xarray.time.values, f"minutes since {first_time}"
                ),
                south_north=("y", np.arange(nexrad_xarray.shape[2])),
                west_east=("x", np.arange(nexrad_xarray.shape[2])),
                projection_x_coordinate=("x", nexrad_xarray.x.values),
                projection_y_coordinate=("y", nexrad_xarray.y.values),
                x=("x", np.arange(nexrad_xarray.shape[2])),
                y=("y", np.arange(nexrad_xarray.shape[2])),
                model_level_number=("z", np.arange(nexrad_xarray.shape[1])),
                altitude=("z", nexrad_xarray.z.values),
            )

            # Adjust dimension names to be standards accepted by iris
            nexrad_xarray["time"] = nexrad_xarray.time.assign_attrs(
                {
                    "standard_name": "time",
                    "long_name": f"minutes since {first_time}",
                    "units": f"minutes since {first_time}",
                }
            )
            nexrad_xarray["z"] = nexrad_xarray.z.assign_attrs(
                {"standard_name": "", "units": "m"}
            )
            nexrad_xarray["lat"] = nexrad_xarray.lat.assign_attrs(
                {"standard_name": "latitude"}
            )
            nexrad_xarray["lon"] = nexrad_xarray.lon.assign_attrs(
                {"standard_name": "longitude"}
            )
            nexrad_xarray["projection_x_coordinate"] = (
                nexrad_xarray.projection_x_coordinate.assign_attrs({"units": "m"})
            )
            nexrad_xarray["projection_y_coordinate"] = (
                nexrad_xarray.projection_y_coordinate.assign_attrs({"units": "m"})
            )
            nexrad_xarray["altitude"] = nexrad_xarray.altitude.assign_attrs(
                {"standard_name": "altitude", "units": "m"}
            )

            return nexrad_xarray

        else:
            raise Exception(
                f"!=====Invalid Tracking Variable. You Entered: {tracking_var.lower()}=====!"
            )

    # If data already grided, just return concated netcdf dataset
    if file_type.lower() == "nc":
        if tracking_var.lower() == "dbz":
            # Read radar objects in and concat into one xarray datarray
            # This is a stupid hacky fix because pyart is dumb
            radar_objects = []
            for file in np.sort(glob.glob(paths_to_files)):
                # Ignore import warnings
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore")
                    radar_objects.append(pyart.io.read_grid(file).to_xarray())

            nexrad_xarray = xr.concat(radar_objects, dim="time").reflectivity
            del radar_objects

            # Subset location and time of interest
            if "multi_nexrad" in CONFIG:
                # Subset time based on user inputs
                if (
                    "min_frame_index" in CONFIG["multi_nexrad"]
                    or "max_frame_index" in CONFIG["multi_nexrad"]
                ):
                    min_frame = (
                        CONFIG["multi_nexrad"]["min_frame_index"]
                        if "min_frame_index" in CONFIG["multi_nexrad"]
                        else 0
                    )
                    max_frame = (
                        CONFIG["multi_nexrad"]["max_frame_index"] + 1
                        if "max_frame_index" in CONFIG["multi_nexrad"]
                        else nexrad_xarray.dims["time"]
                    )

                    nexrad_xarray = nexrad_xarray.isel(
                        time=np.arange(
                            min_frame,
                            max_frame,
                        ),
                        drop=True,
                    )

                if "bounds" in CONFIG["multi_nexrad"]:
                    mask_lon = (
                        nexrad_xarray.lon >= CONFIG["multi_nexrad"]["bounds"][0]
                    ) & (nexrad_xarray.lon <= CONFIG["multi_nexrad"]["bounds"][1])
                    mask_lat = (
                        nexrad_xarray.lat >= CONFIG["multi_nexrad"]["bounds"][2]
                    ) & (nexrad_xarray.lat <= CONFIG["multi_nexrad"]["bounds"][3])

                    nexrad_xarray = nexrad_xarray.where(mask_lon & mask_lat, drop=True)

            else:
                raise Exception("""!=====CONFIG Missing "multi_nexrad" Field=====!""")

            # Replace time dimension with minutes since first time and add other x y z coords
            first_time = nexrad_xarray.time.values[0]
            nexrad_xarray = nexrad_xarray.assign_coords(
                time=cftime.date2num(
                    nexrad_xarray.time.values, f"minutes since {first_time}"
                ),
                south_north=("y", np.arange(nexrad_xarray.shape[2])),
                west_east=("x", np.arange(nexrad_xarray.shape[2])),
                projection_x_coordinate=("x", nexrad_xarray.x.values),
                projection_y_coordinate=("y", nexrad_xarray.y.values),
                x=("x", np.arange(nexrad_xarray.shape[2])),
                y=("y", np.arange(nexrad_xarray.shape[2])),
                model_level_number=("z", np.arange(nexrad_xarray.shape[1])),
                altitude=("z", nexrad_xarray.z.values),
            )

            # Adjust dimension names to be standards accepted by iris
            nexrad_xarray["time"] = nexrad_xarray.time.assign_attrs(
                {
                    "standard_name": "time",
                    "long_name": f"minutes since {first_time}",
                    "units": f"minutes since {first_time}",
                }
            )
            nexrad_xarray["z"] = nexrad_xarray.z.assign_attrs(
                {"standard_name": "", "units": "m"}
            )
            nexrad_xarray["lat"] = nexrad_xarray.lat.assign_attrs(
                {"standard_name": "latitude"}
            )
            nexrad_xarray["lon"] = nexrad_xarray.lon.assign_attrs(
                {"standard_name": "longitude"}
            )
            nexrad_xarray["projection_x_coordinate"] = (
                nexrad_xarray.projection_x_coordinate.assign_attrs({"units": "m"})
            )
            nexrad_xarray["projection_y_coordinate"] = (
                nexrad_xarray.projection_y_coordinate.assign_attrs({"units": "m"})
            )
            nexrad_xarray["altitude"] = nexrad_xarray.altitude.assign_attrs(
                {"standard_name": "altitude", "units": "m"}
            )

            return nexrad_xarray

        raise Exception(
            f"!=====Invalid Tracking Variable. You Entered: {tracking_var.lower()}=====!"
        )

    raise Exception(f"!=====Invalid File Type. You Entered: {file_type.lower()}=====!")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Jul 19 13:12:33 2024

@author: thahn
"""

# =============================================================================
# TODO: Implement all these tests once we can create our test cases
# =============================================================================

import glob
import os
import shelve
import shutil
from pathlib import Path

import geopandas as gpd
import pandas as pd
import xarray as xr
from six.moves import urllib

import CoCoMET


# Returns true if there is a difference between input objects
def diff_output(d1: object, d2: object, path: str = "") -> bool:
    """


    Parameters
    ----------
    d1 : object
        DESCRIPTION.
    d2 : object
        DESCRIPTION.
    path : str, optional
        DESCRIPTION. The default is "".

    Returns
    -------
    bool
        DESCRIPTION.

    """
    for k in d1:
        if k in d2:
            if type(d1[k]) is dict:
                if diff_output(d1[k], d2[k], "%s -> %s" % (path, k) if path else k):
                    return True
            elif type(d1[k]) is tuple:
                for ii in range(len(d1[k])):
                    if (
                        type(d1[k][ii]) is pd.DataFrame
                        or type(d1[k][ii]) is gpd.GeoDataFrame
                    ):
                        if not d1[k][ii].equals(d2[k][ii]):
                            result = [
                                "%s: " % path,
                                " - %s : %s" % (k, d1[k]),
                                " + %s : %s" % (k, d2[k]),
                            ]
                            print("\n".join(result))
                    elif d1[k][ii] != d2[k][ii]:
                        result = [
                            "%s: " % path,
                            " - %s : %s" % (k, d1[k]),
                            " + %s : %s" % (k, d2[k]),
                        ]
                        print("\n".join(result))
                        return True
            elif (
                type(d1[k]) is pd.DataFrame
                or type(d1[k]) is gpd.GeoDataFrame
                or type(d1[k]) is xr.Dataset
            ):
                if not d1[k].equals(d2[k]):
                    result = [
                        "%s: " % path,
                        " - %s : %s" % (k, d1[k]),
                        " + %s : %s" % (k, d2[k]),
                    ]
                    print("\n".join(result))
                    return True
            elif type(d1[k]) is xr.DataArray:
                if not d1[k].broadcast_equals(d2[k]):
                    result = [
                        "%s: " % path,
                        " - %s : %s" % (k, d1[k]),
                        " + %s : %s" % (k, d2[k]),
                    ]
                    print("\n".join(result))
                    return True
            elif type(d1[k]) is CoCoMET.analysis.analysis_object.Analysis_Object:
                if diff_output(
                    d1[k].return_analysis_dictionary(),
                    d2[k].return_analysis_dictionary(),
                    "%s -> %s" % (path, k) if path else k,
                ):
                    return True
            elif d1[k] != d2[k]:
                result = [
                    "%s: " % path,
                    " - %s : %s" % (k, d1[k]),
                    " + %s : %s" % (k, d2[k]),
                ]
                print("\n".join(result))
                return True
        else:
            print("%s%s as key not in d2\n" % ("%s: " % path if path else "", k))

    return False


data_out = Path(os.getcwd())
CONFIG = CoCoMET.CoCoMET_load(str(data_out) + "/CoCoMET/tests/config_for_tests.yml")


def test_load_data():
    # We need to store the data online somewhere
    # Then add a script to download data from online...probably zenodo

    data_out = Path(os.getcwd())
    data_files = glob.glob(str(data_out) + "/.cocomet_testing_datasets/WRF/*")

    if len(data_files) == 0:
        file_path = "https://zenodo.org/records/14611486/files/.cocomet_testing_datasets.zip?download=1"

        temp_zip_file = Path("temp.zip")
        print("=====Downloading and Extracing Data=====")

        urllib.request.urlretrieve(file_path, temp_zip_file)

        shutil.unpack_archive(temp_zip_file, data_out)
        temp_zip_file.unlink()

        data_files = glob.glob(str(data_out) + "/.cococomet_testing_datasets/WRF/*")

    assert True


# """
# TODO: Add full testing coverage
# """


def test_single_core_full_run():
    CONFIG["parallel_processing"] = False

    output = CoCoMET.CoCoMET_start(CONFIG=CONFIG)

    # Get expected output
    with shelve.open(str(data_out) + "/.cocomet_testing_datasets/true_out") as f:
        print(f)
        true_output = f["data"]

    assert diff_output(output, true_output) == False


def test_multi_core_full_run():
    CONFIG["parallel_processing"] = True

    output = CoCoMET.CoCoMET_start(CONFIG=CONFIG)

    # Get expected output
    with shelve.open(str(data_out) + "/.cocomet_testing_datasets/true_out") as f:
        true_output = f["data"]

    assert diff_output(output, true_output) == False



import numpy as np
import pandas as pd
from tqdm import tqdm


def sphericity(
    surface_area_df: pd.DataFrame, volume_df: pd.DataFrame, **args: dict
) -> pd.DataFrame:
    """

    Parameters
    ----------
    surface_area_df : pd.DataFrame
        A pandas dataframe with the 2D surface area values of each cell.
    volume_df : pd.DataFrame
        A pandas dataframe with the volumes of each cell.
    **args : dict
        Throw away variables.

    Returns
    -------
    pandas.core.frame.DataFrame
        A pandas dataframe with the following rows: frame, feature_id, cell_id, sphericity.
    """

    sphericities = []
    frames = []
    feature_ids = []
    cell_ids = []
    for row in tqdm(
        surface_area_df.itertuples(),
        desc="=====Calculating Cell Sphericities=====",
        total=len(surface_area_df),
    ):

        i = row[0]
        frames.append(row[1])
        feature_ids.append(row[2])
        cell_ids.append(row[3])

        surface_area = row[
            4
        ]  # this is assuming both the surface area and volume are in units of km
        volume = volume_df.iloc[i]

        if surface_area == 0 or np.isnan(surface_area):  # check for 0 or NaN values
            sphericities.append(np.NaN)
            continue

        if volume == 0 or np.isnan(volume):  # check for 0 or NaN values
            sphericities.append(np.NaN)
            continue

        sphericity_val = ((np.pi) ** (1 / 3) * (6 * volume) ** (2 / 3)) / surface_area
        if sphericity_val > 1:
            sphericities.append(1)
            continue
        sphericities.append(sphericity_val)

    sphericity_df = pd.DataFrame(
        {
            "frame": frames,
            "feature_id": feature_ids,
            "cell_id": cell_ids,
            "sphericity": sphericities,
        }
    )
    return sphericity_df



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jul 17 14:14:14 2024

@author: thahn
"""

# =============================================================================
# This defines the methods for running MOAAP on WRF data processed using wrf_load.py
# =============================================================================

import os

import numpy as np
import pandas as pd
import xarray as xr

from CoCoMET.MOAAP import moaap

from .mesonh_calculate_products import mesonh_calculate_brightness_temp


# Calculate nearest item in list to given pivot
def find_nearest(array, pivot):
    array = np.asarray(array)
    idx = (np.abs(array - pivot)).argmin()
    return idx


def mesonh_run_moaap(mesonh_xarray: xr.Dataset, CONFIG: dict) -> xr.Dataset:
    """


    Parameters
    ----------
    mesonh_xarray : xarray.core.dataset.Dataset
        xarray Dataset containing MesoNH data calculated from mesonh_load.py.
    CONFIG : dict
        User configuration file.

    Returns
    -------
    mask_file : xarray.core.dataset.Dataset
        Default MOAAP output mask.

    """

    # Get basic setup variables including lat/lon, delta time, a pandas time range vector (TODO: adjust output to )
    latitudes = mesonh_xarray.lat[0].values
    longitudes = mesonh_xarray.lon[0].values
    dt = np.median(np.diff(mesonh_xarray.time).astype("timedelta64[m]")).astype(float)
    times = pd.date_range(
        start=mesonh_xarray.time[0].values,
        end=mesonh_xarray.time[-1].values,
        freq=str(dt) + "min",
    )
    mask = np.ones(latitudes.shape)

    # TODO: Find out how to calculate geopotential heights
    # Get all necessary variables from WRF output to input into MOAAP

    # Get pressure heights
    total_p = mesonh_xarray.P
    avg_geo_pres = [np.mean(h.values) for h in total_p[0]]

    # Get height_idx of 850hPA, 500hPA, and 200hPA
    height_idx_850 = find_nearest(avg_geo_pres, 850)
    # height_idx_500 = find_nearest(avg_geo_pres, 500)
    height_idx_200 = find_nearest(avg_geo_pres, 200)

    # get 850hPA wind speeds and 200hPA wind speeds
    v_winds_850 = mesonh_xarray.v[:, height_idx_850].values
    u_winds_850 = mesonh_xarray.u[:, height_idx_850].values
    v_winds_200 = mesonh_xarray.v[:, height_idx_200].values
    u_winds_200 = mesonh_xarray.u[:, height_idx_200].values

    # Get 850hPA air temperature
    air_temp = mesonh_xarray.T[:, height_idx_850].values

    # Get brightness temp
    tb = (
        mesonh_xarray.TB.values
        if "TB" in mesonh_xarray
        else mesonh_calculate_brightness_temp(mesonh_xarray)
    )

    # Get total mixing ratio at 850hPA
    mr = (
        mesonh_xarray.qv
        + mesonh_xarray.qc
        + mesonh_xarray.qr
        + mesonh_xarray.qi
        + mesonh_xarray.qs
        + mesonh_xarray.qg
    )[:, height_idx_850]

    # Get geopotential heights
    # geopt =

    # Get precipitation rate
    # Convert to per time unit, not per hour
    pr = mesonh_xarray.pcp_rate / (60 / dt)

    # Create output folder if it does not already exist
    if not os.path.exists(CONFIG["mesonh"]["moaap"]["tracking_save_path"]):
        os.makedirs(CONFIG["mesonh"]["moaap"]["tracking_save_path"])

    moaap(
        longitudes,
        latitudes,
        times,
        dt / 60,
        mask,
        DataName="CoMET_MesoNH_MOAAP_TRACKING",
        OutputFolder=CONFIG["mesonh"]["moaap"]["tracking_save_path"],
        # Data Variables
        v850=v_winds_850,
        u850=u_winds_850,
        t850=air_temp,
        q850=mr,
        slp=mesonh_xarray.SLP * 100,
        ivte=None,
        ivtn=None,
        z500=None,
        v200=v_winds_200,
        u200=u_winds_200,
        pr=pr,
        tb=tb,
        # Any user defined params
        **CONFIG["mesonh"]["moaap"],
    )

    output_filepath = (
        CONFIG["mesonh"]["moaap"]["tracking_save_path"]
        + str(times[0].year)
        + str(times[0].month).zfill(2)
        + "_CoMET_MesoNH_MOAAP_TRACKING_ObjectMasks_dt-%.2f" % dt
        + "min_MOAAP-masks"
        + ".nc"
    )
    mask_file = xr.open_mfdataset(
        output_filepath, coords="all", concat_dim="time", combine="nested"
    )

    return mask_file



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jul  3 11:01:10 2024

@author: thahn
"""

# =============================================================================
# This file contains the functions used to calculate additional values from wrf output. Including reflectivity, mass point updrafts, and altitudes
# =============================================================================

import warnings

import numpy as np
import xarray as xr
from tqdm import tqdm


def wrf_calculate_reflectivity(wrf_xarray: xr.Dataset) -> xr.DataArray:
    """


    Parameters
    ----------
    wrf_xarray : xarray.core.dataset.Dataset
        Xarray Dataset containing default WRF values.

    Returns
    -------
    dBZ : xarray.core.dataarray.DataArray
        DataArray containing calculated reflectivity values.

    """

    # Get variables from WRF
    t = wrf_xarray["T"]  # K
    p = wrf_xarray["P"]  # Pa
    pb = wrf_xarray["PB"]  # Pa
    qv = wrf_xarray["QVAPOR"]  # kg kg-1
    qr = wrf_xarray["QRAIN"]  # kg kg-1
    qs = wrf_xarray["QSNOW"]  # kg kg-1
    qg = wrf_xarray["QGRAUP"]  # kg kg-1

    # Calculate proper pressures and actual temperature
    full_t = t + 300  # K
    full_p = p + pb  # Pa

    # c = 2.0/7.0
    # tmk = t * np.power(p * .00001, c) # TODO : check if this is correct, it gives higher dbz values but maybe its not right

    tmk = full_t * (full_p / 1e5) ** (287.0 / 1004.5)

    # Supress divide by zero warnings
    with warnings.catch_warnings():
        warnings.filterwarnings(
            action="ignore", message="divide by zero encountered in divide"
        )

        # Calculate density of dry air at points
        virtual_t = tmk * (0.622 + qv) / (0.622 * (1 + qv))

        dry_air_density = full_p / (287 * virtual_t)

        # Slope intercept constants
        N0r = 8e6
        N0s = 2e7
        N0g = 4e6

        # Calculate individual reflectivites for different types of particles
        gamma_r = ((np.pi * N0r * 1000) / (qr * dry_air_density)) ** (1 / 4)
        Z_er = 720 * N0r * (gamma_r ** (-7)) * 1e18

        gamma_s = ((np.pi * N0s * 100) / (qs * dry_air_density)) ** (1 / 4)
        Z_es = 161.28 * N0s * (gamma_s ** (-7)) * ((100 / 1000) ** 2) * 1e18

        gamma_g = ((np.pi * N0g * 400) / (qg * dry_air_density)) ** (1 / 4)
        Z_eg = 161.28 * N0g * (gamma_g ** (-7)) * ((400 / 1000) ** 2) * 1e18

        # Sum them up
        Z_e = Z_er + Z_es + Z_eg

        # Make sure minimum value is -30dBZ and remove any NaNs (replace NaNs with -30dBZ)
        # Z_e.values = np.clip(Z_e.values, 0.001, 1e99)
        # np.nan_to_num(Z_e.values, copy=False, nan=0.001)

        Z_e = Z_e.clip(0.001, 1e99)
        # np.nan_to_num(Z_e.values, copy=False, nan=0.001)

        dBZ = 10 * np.log10(Z_e)

    # Assign attributes
    dBZ = dBZ.assign_attrs(
        {
            "FieldType": 104,
            "MemoryOrder": "XYZ",
            "description": "radar reflectivity",
            "units": "dBZ",
            "stagger": "",
            "coordinates": "XLONG XLAT XTIME",
        }
    )

    return dBZ.chunk(t.unify_chunks().chunksizes)


def wrf_calculate_brightness_temp(wrf_xarray: xr.Dataset) -> np.ndarray:
    """


    Parameters
    ----------
    wrf_xarray : xarray.core.dataset.Dataset
        Xarray Dataset containing default WRF values.

    Returns
    -------
    TB : numpy.ndarray
        Numpy array containing brightness temperature at each point and time--same dimension as input.

    """

    OLR = wrf_xarray["OLR"].values

    TB = np.empty(OLR.shape)

    a = 1.228
    b = -1.106e-3
    sigma = 5.67e-8  # W m^-2 K^-4

    for tt, ix, iy in tqdm(
        np.ndindex(OLR.shape),
        desc="=====Calculating WRF Brightness Temperatures=====",
        total=np.prod(OLR.shape),
    ):
        tf = (OLR[tt, ix, iy] / sigma) ** 0.25
        TB[tt, ix, iy] = (-a + np.sqrt(a**2 + 4 * b * tf)) / (2 * b)

    return TB


def wrf_calculate_agl_z(wrf_xarray: xr.Dataset) -> xr.DataArray:
    """


    Parameters
    ----------
    wrf_xarray : xarray.core.dataset.Dataset
        Xarray Dataset containing default WRF values.

    Returns
    -------
    geopt : xarray.core.dataarray.DataArray
        Dataarray of heights AGL.

    """

    ph = wrf_xarray["PH"]
    phb = wrf_xarray["PHB"]
    hgt = wrf_xarray["HGT"][0].squeeze().values

    # Make sure we only take one time dimension
    geopt = (ph + phb)[0].squeeze()

    # DESTAGGER geopt
    geopt = 0.5 * geopt[1:] + 0.5 * geopt[:-1]
    geopt = geopt.rename(bottom_top_stag="bottom_top")

    # Account for terrain to convert from MSL to AGL, hence the - hgt
    return (geopt / 9.81) - hgt


def wrf_calculate_wa(wrf_xarray: xr.Dataset) -> xr.DataArray:
    """


    Parameters
    ----------
    wrf_xarray : xarray.core.dataset.Dataset
        Xarray Dataset containing default WRF values.

    Returns
    -------
    wa : xarray.core.dataarray.DataArray
        Dataarray of vertical wind components at mass points.

    """

    # Destagger vertical winds
    wa = (0.5 * wrf_xarray["W"][:, 1:] + 0.5 * wrf_xarray["W"][:, :-1]).rename(
        bottom_top_stag="bottom_top"
    )
    wa = wa.assign_attrs(
        {
            "units": "m s-1",
            "coordinates": "XLONG XLAT XTIME",
            "description": "updraft velocity",
            "MemoryOrder": "XYZ",
        }
    )

    return wa


def wrf_calculate_precip_rate(wrf_xarray: xr.Dataset) -> np.ndarray:
    """


    Parameters
    ----------
    wrf_xarray : xarray.core.dataset.Dataset
        Xarray Dataset containing default WRF values.

    Returns
    -------
    precip_rate : numpy.ndarray
        Numpy array of precipitation rate in mm/hr.

    """

    total_precip = (wrf_xarray.RAINC + wrf_xarray.RAINNC).values
    precip_rate = np.zeros(total_precip.shape)

    for ii in tqdm(
        range(total_precip.shape[0] - 1),
        desc="=====Calculating WRF Precipitation Rate=====",
        total=total_precip.shape[0] - 1,
    ):
        precip_rate[ii] = (total_precip[ii + 1] - total_precip[ii]) * (
            60 / wrf_xarray.DT
        )

    return precip_rate



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Jul 19 13:48:43 2024

@author: thahn
"""

# =============================================================================
# TODO: Create a basic case study of one isolated cell and one MCS to input in the tracks
# Should also convert this one case into the various model types / CoCoMET input types to test.
# This can be used later for unit testing of individual functions
# =============================================================================


def create_isolated_cell(grid_shape, cell_var, cell_radius, max_dbz=None, min_tb=None):
    """

    Parameters
    ----------
    grid_shape : length 4 tuple
        The shape of the domain you want to create in number of grid points (t,x,y,z). i.e. (20,500,500,50).
    cell_var : string, ["dbz", "tb"]
        The variable you want to use to define the cell, either reflectivity or brightness temperature.
    cell_radius : int
        The radius of the cell in # of grid points.
    max_dbz : float, optional
        The maximum desired reflectivity value if using dbz as cell_var. The default is None.
    min_tb : float, optional
        The minimum desired brightness temperature value if using tb as cell_var. The default is None.

    Returns
    -------
    cell_grid : numpy.ndarray
        A numpy array of shape equivelant to grid_shape but has either reflectivity or brightness temperature values.

    """

    import numpy as np
    from scipy.ndimage import gaussian_filter
    from tqdm import tqdm

    cell_grid = np.zeros(grid_shape)

    # Cell should start in corner at mid height and propagate to the center of the grid
    cell_location = [0, 0, int(np.median(np.arange(0, grid_shape[3])))]

    # Determine the number of grid points that need to be moved per frame
    target_point = [int(grid_shape[1] / 2), int(grid_shape[2] / 2)]

    x_steps_per_time = np.ceil(target_point[0] / grid_shape[0])
    y_steps_per_time = np.ceil(target_point[1] / grid_shape[0])

    for tt in tqdm(
        range(grid_shape[0]),
        desc="=====Creating Isolated Cell Grid=====",
        total=grid_shape[0],
    ):
        # Create sphere of radius r
        xx, yy, zz = np.mgrid[: grid_shape[1], : grid_shape[2], : grid_shape[3]]

        sphere = (
            (xx - cell_location[0]) ** 2
            + (yy - cell_location[1]) ** 2
            + (zz - cell_location[2]) ** 2
        )
        sphere = sphere <= (cell_radius**2)

        cell_grid[tt] = sphere

        # Adjust cell position
        cell_location[0] += x_steps_per_time
        cell_location[1] += y_steps_per_time

    # If cell_var is "dbz" set reflectivities
    if cell_var.lower() == "dbz":
        cell_grid[cell_grid == 0] = -30
        cell_grid[cell_grid == 1] = max_dbz

        # Add some filter to make it more cell-like
        for tt in tqdm(
            range(cell_grid.shape[0]),
            desc="=====Filtering Isolated Cell=====",
            total=cell_grid.shape[0],
        ):
            cell_grid[tt] = gaussian_filter(cell_grid[tt], 7)

        return cell_grid

    # Otherwise set brightness temperatures
    cell_grid[cell_grid == 0] = 300
    cell_grid[cell_grid == 1] = min_tb

    # Add some filter to make it more cell-like
    for tt in tqdm(
        range(cell_grid.shape[0]),
        desc="=====Filtering Isolated Cell=====",
        total=cell_grid.shape[0],
    ):
        cell_grid[tt] = gaussian_filter(cell_grid[tt], 5)

    return cell_grid


def create_mcs():
    print("=====In Progress=====")


def create_test_wrf_xarray(cell_grid, dt, dx, dy, dz, cell_var):
    """

    Parameters
    ----------
    cell_grid : numpy.ndarray
        A numpy array of shape (t,x,y,z) but has either reflectivity or brightness temperature values.
    dt : float
        The temporal spacing of the grid in minutes.
    dx : float
        The x spacing of the grid in meters.
    dy : float
        The y spacing of the grid in meters.
    dz : float
        The z spacing of the grid in meters.
    cell_var : string ["dbz", "tb"]
        The variable you want to use to define the cell, either reflectivity or brightness temperature.

    Returns
    -------
    wrf_xarray : xarray.core.dataset.Dataset
        DESCRIPTION.

    """

    import xarray as xr

    print("=====In Progress=====")

    return xr.Dataset()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Jul  8 17:18:18 2024

@author: thahn
"""

# =============================================================================
# This file contains the functions used to calculate additional values from MesoNH output. Including reflectivity and altitudes
# =============================================================================

import warnings

import numpy as np
import xarray as xr
from tqdm import tqdm


def mesonh_calculate_reflectivity(mesonh_xarray: xr.Dataset) -> xr.DataArray:
    """


    Parameters
    ----------
    mesonh_xarray : xarray.core.dataset.Dataset
        Xarray Dataset containing default MesoNH values.

    Returns
    -------
    dBZ : xarray.core.dataarray.DataArray
        DataArray containing calculated reflectivity values.

    """

    # Get variables from MesoNH
    t = mesonh_xarray["T"]
    p = mesonh_xarray["P"]
    qv = mesonh_xarray["qv"]
    qr = mesonh_xarray["qr"]
    qs = mesonh_xarray["qs"]
    qg = mesonh_xarray["qg"]

    # Calculate proper pressures and actual temperature
    full_p = p
    tmk = t

    # Suppress divide by zero warnings
    with warnings.catch_warnings():
        warnings.filterwarnings(
            action="ignore", message="divide by zero encountered in divide"
        )

        # Calculate density of dry air at points
        virtual_t = tmk * (0.622 + qv) / (0.622 * (1 + qv))

        dry_air_density = full_p / (287 * virtual_t)

        # Slope intercept constants
        N0r = 8e6
        N0s = 2e7
        N0g = 4e6

        # Calculate individual reflectivites for different types of particles
        gamma_r = ((np.pi * N0r * 1000) / (qr * dry_air_density)) ** (1 / 4)
        Z_er = 720 * N0r * (gamma_r ** (-7)) * 1e18

        gamma_s = ((np.pi * N0s * 100) / (qs * dry_air_density)) ** (1 / 4)
        Z_es = 161.28 * N0s * (gamma_s ** (-7)) * ((100 / 1000) ** 2) * 1e18

        gamma_g = ((np.pi * N0g * 400) / (qg * dry_air_density)) ** (1 / 4)
        Z_eg = 161.28 * N0g * (gamma_g ** (-7)) * ((400 / 1000) ** 2) * 1e18

        # Sum them up
        Z_e = Z_er + Z_es + Z_eg

        # Make sure minimum value is -30dBZ and remove any NaNs (replace NaNs with -30dBZ)
        Z_e.values = np.clip(Z_e.values, 0.001, 1e99)
        np.nan_to_num(Z_e.values, copy=False, nan=0.001)

        dBZ = 10 * np.log10(Z_e)

    # Assign attributes
    dBZ = dBZ.assign_attrs(
        {
            "FieldType": 104,
            "MemoryOrder": "XYZ",
            "description": "radar reflectivity",
            "units": "dBZ",
            "stagger": "",
            "coordinates": "lon lat time",
        }
    )

    return dBZ.chunk(t.unify_chunks().chunksizes)


def mesonh_calculate_brightness_temp(mesonh_xarray: xr.Dataset) -> np.ndarray:
    """


    Parameters
    ----------
    mesonh_xarray : xarray.core.dataset.Dataset
        Xarray Dataset containing default MesoNH values.

    Returns
    -------
    TB : numpy.ndarray
        Numpy array containing brightness temperature at each point and time--same dimension as input.

    """

    OLR = mesonh_xarray["LWup_TOA"].values

    TB = np.empty(OLR.shape)

    a = 1.228
    b = -1.106e-3
    sigma = 5.67e-8  # W m^-2 K^-4

    for tt, ix, iy in tqdm(
        np.ndindex(OLR.shape),
        desc="=====Calculating MesoNH Brightness Temperatures=====",
        total=np.prod(OLR.shape),
    ):
        tf = (OLR[tt, ix, iy] / sigma) ** 0.25
        TB[tt, ix, iy] = (-a + np.sqrt(a**2 + 4 * b * tf)) / (2 * b)

    return TB


def mesonh_calculate_agl_z(mesonh_xarray: xr.Dataset) -> xr.DataArray:
    """


    Parameters
    ----------
    mesonh_xarray : xr.Dataset
        Xarray Dataset containing default MesoNH values.

    Returns
    -------
    geopt : xarray.core.dataarray.DataArray
        Dataarray of heights AGL in m.

    """

    return mesonh_xarray.Z[0].squeeze()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Jul 11 11:23:21 2024

@author: thahn
"""


# TODO: Refactor this and/or implement 3D calculation. Also implement this for models, not just ARM observations
# NAME: calc_cape.m (v1.0)
#
# PURPOSE: Determine the CAPE, CIN, LCL, LFC, LNB, and W of a parcel using
# parcel theory. User defines how to determine parcel starting properies
# and which processes to include (e.g., freezing, deposition, and
# hydrometeor loading).
#
# INPUT: Requires data that contains pressure, height, temperature, and
# either relative humidity or dew point (Declared in 'Input Data' section).
#
# OUTPUT: Displays the CAPE, CIN, LCL, LFC, LNB, and Wmax of the parcel
# (variables: cape, cin, lcl, lfc, lnb, w_p, respectively). Also generates
# vectors of the parcel sensible temperature and vertical velocity with
# height (variables: temp_log and w_log, respectively).
#
# NOTES:
#   -Can select suface based, mixed-layer, or most unstable parcel defined
#   in 'User Settings' section.
#   -Surface-based parcel defined as: lowest possible data point.
#   -Mixed-layer parcel defined as: parcel with properties of the mean of
#   the mixed-layer that is a user defined depth in 'User Settings'
#   section [Default: 1.5 km]. Parcel starts at surface.
#   -Most-Unstable parcel defined as: parcel that has
#   the greatest Theta-E (using Bolton 1980) below a certain pressure level
#   that is a user defined level in 'User Settings' section [Default:
#   700mb].
#   -Can have pseudoadiabatic/irreversible (i.e., all hydrometeors(water/ice
#   particles formed from condensation) fall out) ascent or moist adiabatic
#   /reversible (i.e,, all hydrometeors are retained with parcel) ascent.
#   pseudoadiabatic meaning
#   -Can have liquid-only or liquid and ice processes. Ice processes
#   include deposition and can include freezing (if loading turned on).
#   -Ice processes are very simplistic, and condensation/deposition rate is
#   based on a linear function of temperature. At 0 deg C: all extra mass
#   resulting from a supersaturated environment is condensed into water. At
#   minus 40 deg C: all extra mass is deposited into ice.
#   -Freezing follows same temperature function as depositon, but only
#   works when liquid hydrometeors are present.
#   -Parcel ascent includes accounting for subsaturated conditions
#   (even though these currently should not happen).
#   -Variables in the code are all defined when first introduced; however,
#   generally variables ending in '_e' represent the environment and '_p'
#   represent the parcel.
#   -Parcel ascent is similar to Bryan and Fritsch (2002).
#
# UPDATE (YYYYMMDD):
#   20180410 - Written Matlab - Mariusz Starzec (M.S.) [v1.0]
#   20190206 - Python 3.6 version - Die Wang
#              read in ARM sounding
#              add in wind shear, midlevel RH, ELR, etc
# % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % %

# Without ice:
# For pseudo-adiabatic (irreversible) ascent, you assume no hydrometeor loading/drag (flag_adiabat = 1, flag_heat =1).
# For moist-adiabatic (reversible) ascent, you retain all the hydrometeors with the parcel (flag_adiabt = 2, flag_heat = 1). This setting essentially retains all hydrometeors and used them to except drag on the parcel. The drag is simplistic, in that it is essentially considered as negative buoyancy where you take the mass of liquid water and apply gravity to it, but it does have a very strong effect.

# With ice:
# For pseudo-adiabatic ascent, only latent heat from deposition (i.e., water vapor to ice) is added in (flag_adiabat = 1, flag_heat = 2).
# For moist adiabatic ascent, there are two options, both include hydrometer loading from both water and ice hydrometeors:
# Only latent heat from deposition is added in (flag_adiabat = 2, flag_heat = 2).
# Latent heat from deposition and freezing of pre-existing liquid water is added in (flag_adiabat = 2, flag_heat = 2, flag_convert = 1). Freezing of pre-existing water can only happen when flag_adiabat = 2, otherwise it has no effect, since you need to carry hydrometeors with you if you want to be able to freeze them. However, this process is overly intense, and ends up releasing a significant amount of heat (which is honestly not as realistic since we end up converting all the hydrometeors to ice which is too much). I really wouldn’t use this option if you are going for realism.

# =========================

import numpy as np
import xarray as xr


def find_closest(A, target):
    # A must be sorted
    idx = A.searchsorted(target)
    idx = np.clip(idx, 1, len(A) - 1)
    left = A[idx - 1]
    right = A[idx]
    idx -= target - left < right - target
    return idx


def calc_tv(t, q):
    """Virt. pot. temp. of environment"""
    return t * (1.0 + 0.61 * q)


def calc_r(e, p):
    """# Water vapor mixing ratio [kg kg-1]"""
    return (0.622 * e) / (p - e)


def calc_es(t):
    """# Vapor pressure - Mangus equation [Pa]"""
    return 610.94 * np.exp((17.625 * t) / (243.04 + t))


def calc_shear(wspd, wdir, alt):
    """wind shear (vector)"""
    index = find_closest(alt, 5000)
    a = wspd[0]
    b = wspd[index]
    c = wdir[0]
    d = wdir[index]
    return (a**2 + b**2 - 2 * a * b * np.cos(np.radians(abs(d - c)))) ** 0.5


def calc_speed_shear(wspd, alt):
    """wind speed shear"""
    a = np.nanmean(wspd[(alt > 0) & (alt < 1000)])
    b = np.nanmean(wspd[(alt > 4000) & (alt < 5000)])
    return b - a


def calc_midlevel_rh(rh, alt):
    return np.nanmean(rh[(alt > 3000) & (alt < 6000)])


def calc_lowlevel_rh(rh, alt):
    return np.nanmean(rh[(alt > 0) & (alt < 3000)])


def calc_richardson(theta_v, u, v, alt):
    """Bulk Richardson Number
    buoyancy / wind shear
    Buoyancy tends to suppress turbulence, while the wind shear tends to generate turbulence mechanically.
    """
    index = find_closest(alt, 6000)
    mask = (alt > 0) & (alt <= 500)
    theta_v0 = np.nanmean(theta_v[mask])
    delta_thetav = theta_v[index] - np.nanmean(theta_v[mask])
    delta_z = alt[index]
    delta_u = u[index] - np.nanmean(u[mask])
    delta_v = v[index] - np.nanmean(v[mask])
    return (9.81 * delta_thetav * delta_z) / (
        np.nanmean(theta_v[mask]) * (delta_u**2 + delta_v**2)
    )


def calc_elr(T, alt):
    # sfc-3km
    ind_3km = find_closest(alt, 3000)
    # find the maximum T st lower altitudes
    ind_max = np.argmax(T[:ind_3km])
    lap_3km = -(T[ind_3km] - T[ind_max]) / (alt[ind_3km] - alt[ind_max]) * 1e3
    # 3km-6km
    ind_6km = find_closest(alt, 6000)
    lap_6km = -(T[ind_6km] - T[ind_3km]) / (alt[ind_6km] - alt[ind_3km]) * 1e3
    return round(lap_3km, 1), round(lap_6km, 1)


# %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% Input Data %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

# Input data (ARM sounding)
# Need Pressure, Height, Temperature and either RH or Dew point.
# Note: Variable names and units need to stay the same. Final units are in
# brackets (e.g., finalized pressure needs to be in Pa, Temp in C, etc.)

# test on the original file
# data = ascii.read('smodel_fine.txt')
# p_e = np.array(data['col1'])
# p_e = p_e*100.   # Pressure [Pa]
# z   = np.array(data['col2']) # Height [m]
# ts_e = np.array(data['col3']) # Temperature [C]
# rh_e = np.array(data['col4'])
# rh_e = rh_e/100               # Relative humidity
# td_e = np.array(data['col4'])


def calculate_interp_sonde_convective_properties(
    sonde: xr.Dataset,
    parcel: int = 3,
    ml_depth: float = 482,
    mu_depth: float = 700.0,
    start: int = 0,
    flag_heat: int = 2,
    flag_convert: int = 0,
    flag_adiabat: int = 1,
    **args: dict,
) -> dict:
    """


    Parameters
    ----------
    sonde : xarray.core.dataset.Dataset
        Input data from INTERPSONDE ARM VAP at one time index.
    parcel : int, optional
        Whether to use 1 - Surface, 2 - Mixed-Layer, or 3 - Most Unstable parcel. The default is 3.
    ml_depth : float, optional
        Depth of mixed-layer [meters] (when parcel = 2). The default is 482.
    mu_depth : float, optional
        Look below this pressure level for MU parcel [hPa] (when parcel = 3). The default is 700.0.
    start : int, optional
        Index to start at for determining parcel properties (e.g. 1 = z(1)).. The default is 0.
    flag_heat : int, optional
        Latent heating only due to liquid or liquid and ice? 1 - Liquid only / 2 - Liquid and Ice. The default is 2.
    flag_convert : int, optional
        Convert pre-existing cloud water to cloud ice? 0 = No / 1 - Yes (using linear function of temperature) Requires flag_adiabat = 2 below (so we have hydrometeors to freeze). The default is 0.
    flag_adiabat : int, optional
        Reversible or Irreversible parcel path? 1 - Pseudoadiabatic (Irreversible) / 2 - Moist Adiabatic (Reversible). The default is 1.
    **args : dict
        Throw away params.

    Raises
    ------
    Exception
        TBD.

    Returns
    -------
    results : dict
        Dictionary of all convective properties (CAPE, CIN, etc.).

    """

    tdryo = sonde["temp"].values  # Dry Bulb Temperature [C]
    p_e = sonde["bar_pres"].values * 1000  # Pressure [Pa]
    rh_eo = sonde["rh"].values  # Relative Humidity [%]
    rh_eo = rh_eo / 100
    z = sonde["height"].values * 1000  # Altitude above mean sea level [m]
    # td_e = ncfile.variables["dp"][:]

    ts_e = tdryo
    rh_e = rh_eo

    u_wind = sonde["u_wind"].values  # eastward wind comp [m/s]
    v_wind = sonde["v_wind"].values  # northward wind comp [m/s]
    wspd = sonde["wspd"].values  # wind speed [m/s]
    wdir = sonde["wdir"].values  # wind direction [degree]

    # %%%%%%%%%%%%%%%%%%%%%%%%%%%%%% User Settings %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

    # Which parcel to use?
    # 1 - Surface / 2 - Mixed-Layer / 3 - Most Unstable
    parcel = parcel
    # depth of mixed-layer [meters] (when parcel = 2)
    ml_depth = ml_depth
    # look below this pressure level for MU parcel [hPa] (when parcel = 3)
    mu_depth = mu_depth

    # Index to start at for determining parcel properties (e.g. 1 = z(1)).
    # This can be used to remove surface data or bad data at low levels.
    # Example: If want to ignore first 4 data levels completely, set start = 5.
    start = start

    # Latent heating only due to liquid or liquid and ice?
    # 1 - Liquid only / 2 - Liquid and Ice
    flag_heat = flag_heat  # To follow Parcel Theory = 1;
    # Convert pre-existing cloud water to cloud ice?
    # 0 = No / 1 - Yes (using linear function of temperature)
    # Requires flag_adiabat = 2 below (so we have hydrometeors to freeze).
    # Can have a very pronounced effect just above the freezing level.
    flag_convert = flag_convert  # To follow Parcel Theory = 0;

    # Reversible or Irreversible parcel path?
    # 1 - Pseudoadiabatic (Irreversible) / 2 - Moist Adiabatic (Reversible)
    flag_adiabat = flag_adiabat  # To follow Parcel Theory = 1;

    # %%%%%%%%%%%%%%%%%%%%%%%%%%% Minor Error Checks %%%%%%%%%%%%%%%%%%%%%%%%%%%%
    # Heights in meters or km? (Should be in meters)
    # if max(z) < 100.:
    #    warning('Expected heights in meters; Assuming data is in kilometers')
    #    z = z*1000.

    # Correct parcel flag?
    # if (parcel>3) | (parcel<1):
    #    error('Incorrect parcel flag')

    # Start index?
    # if start>len(z):
    #    error('Start index exceeds max index')

    # Reasonable depths if ML/MU parcels chosen?
    # if ((parcel==2) & (ml_depth >=max(z))) | ((parcel==3) & (mu_depth >max(z))):
    #    error('ML/MU parcel depths do not match height data')

    # Correct heating flag?
    # if (flag_heat < 1) | (flag_heat > 2):
    #    error('Incorrect heating flag')

    # Correct parcel adiabat?
    # if (flag_adiabat <1) | (flag_adiabat >2):
    #    error('Incorrect adiabat flag')

    # %%%%%%%%%%%%%%%%%%%%%%%%%%%%%% Initialize %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    # Constants
    g = 9.81  # gravity [m s-2]
    cpd = 1004.0  # specific heat of air [J kg-1 K-1]
    ci = 2106.0  # specific heat of ice water [J kg-1 K-1]
    cl = 4186.0  # specific heat of liquid water [J kg-1 K-1]
    cpv = 1885.0  # specific heat of water vapor [J kg-1 K-1]
    Rd = 287.0  # gas constant of dry air [J kg-1 K-1]
    Rv = 461.0  # gas constant of water vapor [J kg-1 K-1]
    eps = Rd / Rv  # epsilon
    lv0 = 2.5e6  # reference latent heat of vaporization [J kg-1]
    lf0 = 3.337e5  # reference latent heat of fusion [J kg-1]
    ls0 = lf0 + lv0  # reference latent heat of sublimation [J kg-1]
    p0 = 1.0e5  # reference pressure [Pa]
    t0 = 273.15  # reference temperature [K]

    # Exner function
    pii = (p_e / p0) ** (Rd / cpd)

    # Environmental Attributes
    # Vapor pressure - Mangus equation [Pa]
    es_e = calc_es(ts_e)

    # If we have RH but not Dew Point
    # This prioritizes RH if both RH and dew point exist since RH is usually
    # the provided/measured variable from soundings and not derived
    # if rh_e.any():
    e_e = rh_e * es_e
    td_e = np.log(e_e / 610.94) * 243.04 / (17.625 - np.log(e_e / 610.94))

    # If we have Dew Point but not RH
    # elif td_e.any():
    # e_e  = calc_es(td_e)
    # rh_e = e_e/es_e

    # Water vapor mixing ratio [kg kg-1]
    r_e = calc_r(e_e, p_e)
    # Saturation Water vapor mixing ratio [kg kg-1]
    rs_e = calc_r(es_e, p_e)
    # Specific humidity [kg kg-1]
    q_e = r_e / (1 + r_e)
    # Saturation Specific humidity [kg kg-1]
    qs_e = rs_e / (1 + rs_e)
    # Convert environmental temperature; [C] to [K]
    ts_e = ts_e + 273.15
    # Pot. temp.
    t_e = ts_e / pii
    # Virt. pot. temp. of environment
    tv_e = calc_tv(t_e, q_e)

    # Get info regarding the starting location/heights of parcel
    if parcel == 1:  # surface parcel
        depth_index = start
    elif parcel == 2:  # mixed-layer parcel
        depth_index = np.where(z <= ml_depth)
        if len(depth_index[0]) < start:
            depth_index[0] = []
        if len(depth_index[0]) < 1:
            raise Exception(
                "error: Mixed-layer depth and starting index are incorrect.",
                " Values result in no starting parcel.",
            )
        elif len(depth_index[0]) == 1:
            raise Exception(
                "warning: Only 1 starting height is valid for mixed-layer parcel"
            )
    elif parcel == 3:  # most unstable parcel
        # Get Theta-E of environment to choose parcel
        # Temperature of lcl (Bolton 1980)
        tlcl_e = 56.0 + (
            1.0 / ((1.0 / ((td_e + t0) - 56.0)) + (np.log(ts_e / (td_e + t0)) / 800.0))
        )
        # Theta-E of the environment (Bolton 1980)
        thetae_e = t_e * np.exp(2.675 * q_e * 1000.0 / tlcl_e)
        thetae_ee = thetae_e[start:]
        depth_index = np.argmax(thetae_ee[p_e[start:] / 100.0 >= mu_depth])
        # If in an extrememly unlikely situation we end up with two heights
        # having the same theta-E, choose the higher one.
        # depth_index = max(depth_index)
        start = depth_index

    # Boolean vector holding starting parcel locations (1 = Valid)
    temp_z = np.zeros(len(z))
    temp_z[depth_index] = 1

    # Starting Parcel Attributes
    # Pot. temp. of parcel
    t_p = t_e * temp_z
    t_p[t_p == 0] = np.nan
    t_p = np.nanmean(t_p)

    # Water vapor mixing ratio of parcel
    q_p = q_e * temp_z
    q_p[q_p == 0] = np.nan
    q_p = np.nanmean(q_p)

    # Virt. pot. temp. of parcel
    tv_p = calc_tv(t_p, q_p)
    flag1 = 0

    # %%%%%%%%%%%%%%%%%%%%%%%%%%%%%% Parcel Ascent %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    # Preallocate Arrays
    # (buoyancy, cape, cin)
    B = np.zeros(len(z) - (start))
    cape = np.zeros(len(B))
    cin = np.zeros(len(B))

    # Initialize values
    # (lcl, lfc, parcel vertical velocity, liquid water, ice water)
    lcl = 9999.9
    lfc = 9999.9
    q_l = 0.0
    q_i = 0.0

    # Logging: Track parcel ascent sensible temperature and vertical velocity
    temp_log = np.empty(len(z) - (start)) * np.nan
    w_log = np.empty(len(z) - (start)) * np.nan

    # Starting parcel values
    temp_log[start] = t_p * pii[start]
    w_log[start] = 0  # assuming parcel is starting with no momentum

    # Raise parcel and determine buoyancy
    for k in range(start + 1, len(z) - start):
        # Current sensible temperature of parcel
        ts_p = t_p * pii[k]

        # Current latent heat of vaporization (based in parcel temperature)
        # Bryan and Fritsch (2002)
        lv_p = lv0 - (cl - cpv) * (ts_p - t0)
        ls_p = ls0 - (ci - cpv) * (ts_p - t0)

        # Current saturation mixing ratio of liquid
        # Magnus equation for saturation vapor pressure
        es_p = calc_es(ts_p - t0)  # vapor pres.
        rs_p = calc_r(es_p, p_e[k])  # mixing ratio
        qs_p = rs_p / (1 + rs_p)  # specific humidity

        # Ice saturation mixing ratio
        esi_p = 611.2 * np.exp(21.8745584 * (ts_p - t0) / (ts_p - 7.66))  # vapor pres.
        rsi_p = eps * esi_p / (p_e[k] - esi_p)  # mixing ratio
        qsi_p = rsi_p / (1 + rsi_p)  # specific humidity

        # % % % FOR LIQUID PROCESSES ONLY % % %
        # Adjust by latent heat of vaporization by condensing in new mass when
        # parcel is supersaturated
        # Rate of conversion from Bryan and Fritsch (2002)
        if flag_heat == 1:
            # % IRREVERSIBLE ASCENT % %
            # No hydrometeors
            delqvl = 0
            if q_p > qs_p:
                lcl = min(lcl, z[k] / 1e3)  # LCL height [km]
                delqvl = (q_p - qs_p) / (1.0 + (qs_p * lv_p**2) / (cpd * Rv * ts_p**2))
                t_p = t_p + (delqvl * lv_p / cpd) / pii[k]
                q_p = q_p - delqvl

            # % % REVERSIBLE ASCENT % %
            # % Keep track of hydrometeors mass
            if flag_adiabat == 2:
                q_l = q_l + delqvl
                # Allow for evaporation of hydrometers if for some unlikely
                # reason we do end up subsaturated
                if q_p < qs_p:
                    delqvl = min(q_l, abs(q_p - qs_p)) / (
                        1.0 + (qs_p * lv_p**2) / (cpd * Rv * ts_p**2)
                    )
                    t_p = t_p - (delqvl * lv_p / cpd) / pii[k]
                    q_l = q_l - delqvl
                    q_p = q_p + delqvl

        # % % % FOR LIQUID AND ICE PROCESSES % % %
        # % Where parcel is saturated: adjust by latent heat of condensation
        if flag_heat == 2:
            # % % IRREVERSIBLE ASCENT % %
            # No hydrometeors
            # The amount of vapor to convert to water or ice is based on a
            # linear function of temperature. All ice when T <= -40 deg C and
            # all water when T >= 0 deg C.
            ratio = max(min((ts_p - 233.15) / (t0 - 233.15), 1.0), 0.0)
            delq = min(q_p, ratio * qs_p + (1.0 - ratio) * qsi_p)
            ice_gain = max((1.0 - ratio) * (q_p - delq), 0.0)
            liq_gain = max(q_p - delq - ice_gain, 0.0)
            if (liq_gain > 0) and (lcl == 9999.9):
                lcl = min(lcl, (z[k] / 1e3))

            delqvi = (ice_gain) / (1.0 + (qsi_p * ls_p**2) / (cpd * Rv * ts_p**2))
            delqvl = (liq_gain) / (1.0 + (qs_p * lv_p**2) / (cpd * Rv * ts_p**2))
            t_p = t_p + ((delqvi * ls_p + delqvl * lv_p) / cpd) / (pii[k])
            q_p = q_p - (delqvi + delqvl)

            # % REVERSIBLE ASCENT % %
            # Keep track of hydrometeors mass
            if flag_adiabat == 2:
                # % % % FREEZING OF PRE-EXISTING WATER % % %
                if flag_convert == 1:
                    # Convert a fraction of water to ice based on the same
                    # linear temperature function.
                    ratio = max(min((ts_p - 233.15) / (t0 - 233.15), 1.0), 0.0)
                    liq_conv = min(q_l, (1.0 - ratio) * q_l)
                    t_p = t_p + ((liq_conv * (ls_p - lv_p)) / cpd) / (pii[k])
                else:
                    liq_conv = 0

                q_l = q_l + delqvl - liq_conv
                q_i = q_i + delqvi + liq_conv

                # Allow for evaporation of hydrometers if for some unlikely reason we do end up subsaturated.
                # Ratio of how much ice loss
                # should occur relative to liquid follows similar linear
                # temperature curve (but inverse, so more liquid loss at cold temps)
                ratio = max(min((ts_p - 233.15) / (t0 - 233.15), 1.0), 0.0)
                delq = max(q_p, ratio * qs_p + (1.0 - ratio) * qsi_p)
                ice_loss = min(q_i, abs((1.0 - ratio) * (q_p - delq)))
                liq_loss = min(q_l, abs(q_p - delq + ice_loss))
                ice_loss = min(q_i, ice_loss + (delq - q_p - liq_loss - ice_loss))
                delqvi = (ice_loss) / (1.0 + (qsi_p * ls_p**2) / (cpd * Rv * ts_p**2))
                delqvl = (liq_loss) / (1.0 + (qs_p * lv_p**2) / (cpd * Rv * ts_p**2))
                t_p = t_p - ((delqvi * ls_p + delqvl * lv_p) / cpd) / (pii[k])
                q_p = q_p + (delqvi + delqvl)
                q_l = q_l - delqvl
                q_i = q_i - delqvi

        # Get new adjusted virt. pot. temp. of parcel
        tv_p = t_p * (1.0 + 0.61 * q_p - (q_l + q_i))
        # Determine Buoyancy
        B[k] = g * ((tv_p - tv_e[k]) / tv_e[k])
        # Determine CAPE/CIN [J kg-1]
        cape[k] = B[k] * (z[k] - z[k - 1]) * (B[k] > 0)
        cin[k] = B[k] * (z[k] - z[k - 1]) * (B[k] < 0)

        # Get new sensible temperature
        # Note: A sort of "equivalent" sensible temperature of parcel: the
        # temperature of the parcel if all effects (such as loading) were
        # accounted for and represented only as a change in temperature
        ts_p = (tv_p / (1.0 + 0.61 * q_p)) * pii[k]

        # Get "traditional" temperature-based LFC if LCL exists
        # This will pick up an embedded pockets of instability as the LFC if
        # there is such a pocket present.
        if (lcl < 9999.9) and (ts_p >= ts_e[k]):
            lfc = min(lfc, z[k] / 1.0e3)  # [km]
            # Alternative LFC: Mean between two points, 1st unstable and last
            # stable point. Probably more "accurate" representation of actual
            # LFC if the grid is coarse.
            # lfc = min(lfc,((z(k)+z(k-1))/2)/1e3);

        # Save logs of parcel sensible temperature and vertical velocity
        temp_log[k] = ts_p  # [K]
        w_log[k] = np.sqrt(2.0 * np.nansum(cape[start + 1 : k + 1])) - np.sqrt(
            2 * np.nansum(abs(cin[start + 1 : k + 1]))
        )  # [m/s]

    # Determine LNB [km]: Height where parcel no longer buoyant.
    # Look for: 1st negatively buoyany height with no positive buoyancy above
    lnb = np.nan  # [km]
    for k in range(start + 1, len(z) - start):
        # print (k,lnb,z[k-1]/1000.)
        if (B[k] < 0) & (B[k - 1] > 0) & (np.nansum(B[k:] > 0) == 0):
            hgt = z[k - 1] / 1.0e3
            lnb = np.nanmin([hgt, lnb])
            if lnb == 0:
                lnb = np.nan

    # "Trim" CAPE/CIN to include everything below LNB only and get one value
    # cape = sum(cape.*(z/1e3 < lnb));
    CIN = np.nansum(cin[z[: len(z) - start] / 1.0e3 < lnb])

    # Alternative CAPE: Only positive buoyancy between LFC and LNB. This
    # ignores pockets of instability between parcel starting point and the LFC.
    # Would eliminate CAPE resulting from surface "superadiabatic" or dry
    # convective CAPE.
    CAPE = np.nansum(
        cape[(z[: len(z) - start] / 1.0e3 < lnb) & (z[: len(z) - start] / 1.0e3 >= lfc)]
    )
    # low-level cape
    LCAPE = np.nansum(
        cape[
            (z[: len(z) - start] / 1.0e3 < lfc + 4)
            & (z[: len(z) - start] / 1.0e3 >= lfc)
        ]
    )

    # print ('lfc=', round(ZLFc,0))
    NCAPE = CAPE / (lnb * 1.0e3 - lfc * 1.0e3)

    # Vertical Velocity of Parcel
    w_p = np.sqrt(2 * np.nansum(CAPE)) - np.sqrt(2 * np.nansum(abs(CIN)))

    # wind shear
    wshr = calc_shear(wspd, wdir, z)
    wsshr = 0.0

    # midlevel humidity
    if rh_e[(z > 3000) & (z < 6000)].shape[0] == 0:
        mid_rh = np.nan
    else:
        mid_rh = calc_midlevel_rh(rh_e, z)

    # lowlevel humidity
    low_rh = calc_lowlevel_rh(rh_e, z)

    # Richardson number
    richardson = calc_richardson(tv_e, u_wind, v_wind, z)

    # calculate the average environmental lapse rate
    try:
        elr3, elr6 = calc_elr(ts_e, z)
    except:
        elr3 = np.nan
        elr6 = np.nan

    results = {
        "CAPE": CAPE,
        "CIN": CIN,
        "LNB": lnb,
        "LFC": lfc,
        "LCL": lcl,
        "Wind_Shear": wshr,
        "Low_RH": low_rh,
        "Mid_RH": mid_rh,
        "Richardson": richardson,
        "ELR_0-3km": elr3,
        "NCAPE": NCAPE,
    }

    return results



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Aug 26 14:40:33 2024

@author: hweiner
"""

# =============================================================================
# Takes in a filepath containing RAMS netCDF data and converts it to a netcdf dataset and/or an iris cube for use in trackers
# =============================================================================

import logging

import cftime
import iris.cube
import matplotlib.pyplot as plt
import numpy as np
import xarray as xr

from .rams_calculate_products import (
    rams_calculate_brightness_temp,
    rams_calculate_precip_rate,
    rams_calculate_reflectivity,
    rams_calculate_wa,
)
from .rams_configure import configure_rams
from .ramscube import load


# TODO: You should be able to track on any variable in the RAMS dataset, also need to add PR tracking
def rams_load_netcdf_iris(
    filepath: str,
    tracking_var: str,
    path_to_header: str,
    CONFIG: dict = None,
) -> tuple[iris.cube.Cube, xr.Dataset]:
    """


    Parameters
    ----------
    filepath : str
        glob style path to rams files (i.e. ./data/ramsout/ramsout_d03_*.h5).
    tracking_var : str
        ["dbz","tb","wa"], variable which is going to be used for tracking--either reflectivity, brightness temperature, or updraft velocity.
    path_to_header : str
        glob style path to rams header files (i.e. ./data/ramsout/ramsheader_*.txt).
    CONFIG : dict, optional
        Standard CoCoMET CONFIG file. The default is None.
    debug : int, optional
        Debug verbosity level, from 0-2. The default is 0.

    Raises
    ------
    Exception
        DESCRIPTION.

    Returns
    -------
    cube : iris.cube.Cube
        iris cube containing either reflectivity, updraft velocity, or brightness temperature values.
    rams_xarrat : xarray.core.dataset.Dataset
        Xarray dataset containing merged rams data.

    """

    rams_xarray = xr.open_mfdataset(
        filepath,
        coords="all",
        concat_dim="Time",
        combine="nested",
        decode_times=False,
        engine="h5netcdf",
        phony_dims="sort",
    )

    # This is already in configure_rams, but eventually we will call RAMS-MAT. Then either put the time formatting here or keep it in RAMS-MAT
    if "rams" in CONFIG:
        # Check for idealized data then correct times
        if "is_idealized" in CONFIG["rams"]:
            if CONFIG["rams"]["is_idealized"]:
                # Update Times
                # Get time differences in minutes
                time_diffs = (
                    np.diff(rams_xarray.Times.values)
                    .astype("timedelta64[m]")
                    .astype("int")
                )

                time_list = [0]

                for diff in time_diffs:
                    time_list.append(time_list[-1] + diff)

                # Initialize simulation at January 1, 2000 for convinence
                time_list = cftime.num2date(
                    time_list, units="minutes since 2000-01-01 00:00:00"
                )
                rams_xarray.assign_coords(XTIME=("Time", time_list))
                rams_xarray["XTIME"] = rams_xarray["XTIME"].assign_attrs(
                    {"description": "minutes since 2000-01-01 00:00:00"}
                )

        # If there are specified bounds, bound the data
        if "bounds" in CONFIG["rams"]:
            # If it is idealized data, print a warning
            if "is_idealized" in CONFIG["rams"]:
                if CONFIG["rams"]["is_idealized"]:
                    logging.warning("!=====Setting bounds for idealized data=====!")

            mask_lon = (rams_xarray.GLON >= CONFIG["rams"]["bounds"][0]) & (
                rams_xarray.GLON <= CONFIG["rams"]["bounds"][1]
            )
            mask_lat = (rams_xarray.GLAT >= CONFIG["rams"]["bounds"][2]) & (
                rams_xarray.GLAT <= CONFIG["rams"]["bounds"][3]
            )

            rams_xarray = rams_xarray.where(mask_lon & mask_lat, drop=True)

        # Subset time based on user inputs
        if "min_frame_index" in CONFIG["rams"] or "max_frame_index" in CONFIG["rams"]:
            min_frame = (
                CONFIG["rams"]["min_frame_index"]
                if "min_frame_index" in CONFIG["rams"]
                else 0
            )
            max_frame = (
                CONFIG["rams"]["max_frame_index"] + 1
                if "max_frame_index" in CONFIG["rams"]
                else rams_xarray.dims["Time"]
            )

            rams_xarray = rams_xarray.isel(
                Time=np.arange(
                    min_frame,
                    max_frame,
                ),
                drop=True,
            )

    else:
        raise Exception("""!=====CONFIG Missing "rams" Field=====!""")

    if tracking_var.lower() == "tb":

        # Configure rams xarray for brightness temperature
        rams_xarray = configure_rams(
            rams_xarray,
            path_to_header,
            CONFIG=CONFIG,
            configure_variables=["TOA_OLR", "LWUP"],
        )

        # Brightness temperature is only 2d so no heights needed
        rams_xarray["TB"] = rams_calculate_brightness_temp(rams_xarray)
        rams_xarray["TB"].attrs["units"] = "K"

        cube = load(rams_xarray, "TB")

    elif tracking_var.lower() == "pr":

        use_available_variables = False
        # Configure rams xarray for brightness temperature

        # If there is a specified calculation type for precipitation, use that
        if "calculation_type" in CONFIG["rams"]:
            if (
                CONFIG["rams"]["calculation_type"]
                == "surface time averaged precipitation rate"
                or CONFIG["rams"]["calculation_type"] is None
            ):
                rams_xarray = configure_rams(
                    rams_xarray,
                    path_to_header,
                    CONFIG=CONFIG,
                    configure_variables=[
                        "ACCPR",
                        "ACCPP",
                        "ACCPS",
                        "ACCPA",
                        "ACCPG",
                        "ACCPH",
                        "ACCPD",
                        "ACONPR",
                        "ACCPIP",
                        "ACCPIC",
                        "ACCPID",
                    ],
                )
                rams_xarray["PR"] = rams_calculate_precip_rate(
                    rams_xarray, calculation_type=CONFIG["rams"]["calculation_type"]
                )

            elif (
                CONFIG["rams"]["calculation_type"]
                == "surface instantaneous precipitation rate"
            ):
                rams_xarray = configure_rams(
                    rams_xarray,
                    path_to_header,
                    CONFIG=CONFIG,
                    configure_variables=[
                        "PCPRR",
                        "PCPRP",
                        "PCPRS",
                        "PCPRA",
                        "PCPRG",
                        "PCPRH",
                        "PCPRD",
                        "CONPRR",
                        "PCPRIP",
                        "PCPRIC",
                        "PCPRID",
                    ],
                )
                rams_xarray["PR"] = rams_calculate_precip_rate(
                    rams_xarray, calculation_type=CONFIG["rams"]["calculation_type"]
                )

            elif (
                CONFIG["rams"]["calculation_type"]
                == "volumetric instantaneous precipitation rate"
            ):
                rams_xarray = configure_rams(
                    rams_xarray,
                    path_to_header,
                    CONFIG=CONFIG,
                    configure_variables=[
                        "PCPVR",
                        "PCPVP",
                        "PCPVS",
                        "PCPVA",
                        "PCPVG",
                        "PCPVH",
                        "PCPVD",
                        "CONPRR",
                        "PCPVIP",
                        "PCPVIC",
                        "PCPVID",
                    ],
                )
                rams_xarray["PR"] = rams_calculate_precip_rate(
                    rams_xarray, calculation_type=CONFIG["rams"]["calculation_type"]
                )

            else:
                print(
                    "No calculation type found in CONFIG file, using available variables"
                )
                use_available_variables = True

        # If there is no calculation type, see which variables are available and use those
        if "calculation_type" not in CONFIG["rams"] or use_available_variables:
            if "ACCPR" in rams_xarray:
                rams_xarray = configure_rams(
                    rams_xarray,
                    path_to_header,
                    CONFIG=CONFIG,
                    configure_variables=[
                        "ACCPR",
                        "ACCPP",
                        "ACCPS",
                        "ACCPA",
                        "ACCPG",
                        "ACCPH",
                        "ACCPD",
                        "ACONPR",
                        "ACCPIP",
                        "ACCPIC",
                        "ACCPID",
                    ],
                )
                rams_xarray["PR"] = rams_calculate_precip_rate(
                    rams_xarray,
                    calculation_type="surface time averaged precipitation rate",
                )

            elif "PCPRR" in rams_xarray:
                rams_xarray = configure_rams(
                    rams_xarray,
                    path_to_header,
                    CONFIG=CONFIG,
                    configure_variables=[
                        "PCPRR",
                        "PCPRP",
                        "PCPRS",
                        "PCPRA",
                        "PCPRG",
                        "PCPRH",
                        "PCPRD",
                        "CONPRR",
                        "PCPRIP",
                        "PCPRIC",
                        "PCPRID",
                    ],
                )
                rams_xarray["PR"] = rams_calculate_precip_rate(
                    rams_xarray,
                    calculation_type="surface instantaneous precipitation rate",
                )

            elif "PCPVR" in rams_xarray:
                rams_xarray = configure_rams(
                    rams_xarray,
                    path_to_header,
                    CONFIG=CONFIG,
                    configure_variables=[
                        "PCPVR",
                        "PCPVP",
                        "PCPVS",
                        "PCPVA",
                        "PCPVG",
                        "PCPVH",
                        "PCPVD",
                        "CONPRR",
                        "PCPVIP",
                        "PCPVIC",
                        "PCPVID",
                    ],
                )
                rams_xarray["PR"] = rams_calculate_precip_rate(
                    rams_xarray,
                    calculation_type="volumetric surface instantaneous precipitation rate",
                )

        rams_xarray["PR"].attrs["units"] = "mm/hr"

        cube = load(rams_xarray, "PR")

        # If tracking 3D precipitation rate, add altitude to the cube
        if len(rams_xarray["PR"].values.shape) == 4:
            cube.coord("altitude").points = rams_xarray["altitudes"].values

    elif tracking_var.lower() == "dbz":

        # Configure rams xarray for reflectivity
        rams_xarray = configure_rams(
            rams_xarray,
            path_to_header,
            CONFIG=CONFIG,
            configure_variables=[
                "RRP",
                "RPP",
                "RSP",
                "RAP",
                "RGP",
                "RHP",
                "CRP",
                "CPP",
                "CSP",
                "CAP",
                "CGP",
                "CHP",
            ],
        )

        rams_xarray["DBZ"] = rams_calculate_reflectivity(rams_xarray)

        cube = load(rams_xarray, "DBZ")
        cube.coord("altitude").points = rams_xarray[
            "altitudes"
        ].values  # there is already an altitude coordinate in the xarray

        # Add altitude field for easier processing later
        rams_xarray["DBZ"] = rams_xarray["DBZ"].assign_coords(
            altitude=(["bottom_top"], rams_xarray["altitudes"].values)
        )

    elif tracking_var.lower() == "wa":

        # Configure rams xarray for brightness temperature
        rams_xarray = configure_rams(
            rams_xarray, path_to_header, CONFIG=CONFIG, configure_variables=[]
        )

        # Get updraft velocity at mass points
        rams_wa = rams_calculate_wa(rams_xarray)

        rams_xarray["WA"] = rams_wa

        cube = load(rams_xarray, "WA")
        cube.coord("altitude").points = rams_xarray[
            "altitudes"
        ].values  # there is already an altitude coordinate in the xarray

        # Add altitude field for easier processing later
        rams_xarray["WA"] = rams_xarray["WA"].assign_coords(
            altitude=("bottom_top", rams_xarray["altitudes"].values)
        )

    else:
        # If not any of the above, try using user inputed value
        try:
            # Configure rams xarray for brightness temperature
            rams_xarray = configure_rams(
                rams_xarray,
                path_to_header,
                CONFIG=CONFIG,
                configure_variables=[tracking_var.upper()],
            )

            var_values = rams_xarray[tracking_var.upper()]
            cube = load(rams_xarray, tracking_var.upper())

            if len(var_values.shape) == 4:
                # Add correct altitude based off of average height at each height index

                cube.coord("altitude").points = rams_xarray[
                    "altitudes"
                ].values  # there is already an altitude coordinate in the xarray

                # Add altitude field for easier processing later
                rams_xarray[tracking_var.upper()] = rams_xarray[
                    tracking_var.upper()
                ].assign_coords(
                    altitude=("bottom_top", rams_xarray["altitudes"].values)
                )

        except:
            raise Exception(
                f"!=====Invalid Tracking Variable. You Entered: {tracking_var.lower()}=====!"
            )

    return (cube, rams_xarray.unify_chunks())
